from __future__ import annotations

import argparse
import calendar
import copy
import datetime
import fnmatch
import json
import os.path
import re
from urllib.parse import urlparse, urlunparse
from typing import Any, List, Optional, Dict, Union, Tuple, Set, Pattern

import requests
from cryptography import x509
from cryptography.hazmat.primitives import serialization

from ... import crypto, constants, utils, api
from ...params import KeeperParams
from ...pedm import admin_plugin, pedm_shared, admin_types, admin_storage
from ...proto import NotificationCenter_pb2, pedm_pb2
from .. import base
from ..helpers import report_utils, prompt_utils, whoami
from . import pedm_aram


class PedmUtils:
    @staticmethod
    def resolve_single_agent(pedm: admin_plugin.PedmPlugin, agent_uid: Any) -> admin_types.PedmAgent:
        if not isinstance(agent_uid, str):
            raise base.CommandError(f'Invalid agent_name: {agent_uid}')

        agent = pedm.agents.get_entity(agent_uid)
        if agent:
            return agent
        raise base.CommandError(f'Agent UID \"{agent_uid}\" does not exist')

    @staticmethod
    def resolve_single_deployment(pedm: admin_plugin.PedmPlugin, deployment_name: Any) -> admin_types.PedmDeployment:
        if not isinstance(deployment_name, str):
            raise base.CommandError(f'Invalid deployment name: {deployment_name}')

        deployment = pedm.deployments.get_entity(deployment_name)
        if deployment:
            return deployment

        l_deployment_name = deployment_name.lower()
        deployments = [x for x in pedm.deployments.get_all_entities() if x.name.lower() == l_deployment_name]
        if len(deployments) == 0:
            raise base.CommandError(f'Deployment \"{deployment_name}\" does not exist')
        if len(deployments) >= 2:
            raise base.CommandError(f'Deployment \"{deployment_name}\" is not unique. Please use Deployment UID')

        return deployments[0]

    @staticmethod
    def resolve_existing_policies(pedm: admin_plugin.PedmPlugin, policy_names: Any) -> List[admin_types.PedmPolicy]:
        found_policies: Dict[str, admin_types.PedmPolicy] = {}
        p: Optional[admin_types.PedmPolicy]
        if isinstance(policy_names, list):
            for policy_name in policy_names:
                p = pedm.policies.get_entity(policy_name)
                if p is None:
                    raise base.CommandError(f'Policy name "{policy_name}" is not found')
                found_policies[p.policy_uid] = p
        if len(found_policies) == 0:
            raise base.CommandError('No policies were found')
        return list(found_policies.values())

    @staticmethod
    def resolve_single_policy(pedm: admin_plugin.PedmPlugin, policy_uid: Any) -> admin_types.PedmPolicy:
        if not isinstance(policy_uid, str):
            raise base.CommandError(f'Invalid policy UID: {policy_uid}')
        policy = pedm.policies.get_entity(policy_uid)

        if isinstance(policy, admin_types.PedmPolicy):
            return policy
        raise base.CommandError(f'Policy UID \"{policy_uid}\" does not exist')

    @staticmethod
    def get_collection_name_lookup(
            pedm: admin_plugin.PedmPlugin
    ) -> Dict[str, Union[admin_types.PedmCollection, List[admin_types.PedmCollection]]]:
        collection_lookup: Dict[str, Union[admin_types.PedmCollection, List[admin_types.PedmCollection]]] = {}

        for collection in pedm.collections.get_all_entities():
            if not isinstance(collection.collection_data, dict):
                continue
            collection_name = collection.collection_data.get('Name')
            if not isinstance(collection_name, str) and not collection_name:
                continue
            collection_name = collection_name.lower()
            collection_lookup[collection_name] = collection
            c = collection_lookup.get(collection_name)
            if c is None:
                collection_lookup[collection_name] = collection
            elif isinstance(c, list):
                c.append(collection)
            elif isinstance(c, admin_types.PedmCollection):
                collection_lookup[collection_name] = [c, collection]
        return collection_lookup

    @staticmethod
    def get_orphan_resources(pedm: admin_plugin.PedmPlugin) -> List[str]:
        resource_types = {pedm_shared.CollectionType.OsBuild, pedm_shared.CollectionType.Application,
                          pedm_shared.CollectionType.UserAccount, pedm_shared.CollectionType.GroupAccount}
        collections = {x.collection_uid for x in pedm.storage.collections.get_all_entities() if x.collection_type in resource_types}
        links = {x.collection_uid for x in pedm.storage.collection_links.get_all_links() if x.link_type == pedm_pb2.CollectionLinkType.CLT_AGENT}
        return list(collections.difference(links))

    @staticmethod
    def resolve_existing_collections(
            pedm: admin_plugin.PedmPlugin,
            collection_names: Any,
            *,
            collection_type: Optional[int] = None,
            ignore_missing: bool = False,
    ) -> List[admin_types.PedmCollection]:

        found_collections: Dict[str, admin_types.PedmCollection] = {}
        if not isinstance(collection_names, list):
            collection_names = [collection_names]

        resolve_by_name = []
        for name in collection_names:
            if not isinstance(name, str) and not ignore_missing:
                raise base.CommandError(f'Invalid collection name: {name}')

            collection = pedm.collections.get_entity(name)
            if collection is None:
                resolve_by_name.append(name)
            else:
                found_collections[collection.collection_uid] = collection

        if len(resolve_by_name) > 0:
            collection_lookup = PedmUtils.get_collection_name_lookup(pedm)
            for name in resolve_by_name:
                c: Optional[admin_types.PedmCollection] = None
                cc = collection_lookup.get(name)
                if cc is None:
                    cc = collection_lookup.get(name.lower())
                if isinstance(cc, admin_types.PedmCollection):
                   c = cc
                elif isinstance(cc, list):
                    if len(cc) > 1 and isinstance(collection_type, int):
                        cc = [x for x in cc if x.collection_type == collection_type]
                    if len(cc) == 1:
                        c = cc[0]
                    else:
                        if not ignore_missing:
                            raise base.CommandError(f'Collection \"{name}\" is not unique. Please use Collection UID')
                if c is None:
                    if not ignore_missing:
                        raise base.CommandError(f'Collection "{name}" is not found')
                else:
                    found_collections[c.collection_uid] = c
        return list(found_collections.values())


class PedmCommand(base.GroupCommandNew):
    def __init__(self):
        super().__init__('Privilege Manager - PEDM')
        self.register_command_new(PedmSyncDownCommand(), 'sync-down')
        self.register_command_new(PedmDeploymentCommand(), 'deployment', 'd')
        self.register_command_new(PedmAgentCommand(), 'agent', 'a')
        self.register_command_new(PedmPolicyCommand(), 'policy', 'p')
        self.register_command_new(PedmCollectionCommand(), 'collection', 'c')
        self.register_command_new(PedmApprovalCommand(), 'approval')
        self.register_command_new(pedm_aram.PedmReportCommand(), 'report')
        #self.register_command_new(PedmBICommand(), 'bi')


class PedmSyncDownCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='sync-down', description='Sync down PEDM data from the backend')
        parser.add_argument('--reload', dest='reload', action='store_true', help='Perform full sync')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs):
        plugin = admin_plugin.get_pedm_plugin(context)
        plugin.sync_down(reload=kwargs.get('reload') is True)


class PedmDeploymentCommand(base.GroupCommandNew):
    def __init__(self):
        super().__init__('Manage PEDM deployments')
        self.register_command_new(PedmDeploymentListCommand(), 'list', 'l')
        self.register_command_new(PedmDeploymentAddCommand(), 'add', 'a')
        self.register_command_new(PedmDeploymentUpdateCommand(), 'edit')
        self.register_command_new(PedmDeploymentDeleteCommand(), 'delete')
        self.register_command_new(PedmDeploymentDownloadCommand(), 'download')
        self.default_verb = 'list'


class PedmDeploymentListCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='list', description='List PEDM deployments', parents=[base.report_output_parser])
        parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                            help='print verbose information')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs):
        plugin = admin_plugin.get_pedm_plugin(context)

        verbose = kwargs.get('verbose') is True
        table: List[List[Any]] = []
        headers = ['deployment_uid', 'name', 'disabled', 'created', 'updated']
        if verbose:
            headers.append('agents')
        else:
            headers.append('agent_count')
        row: List[Any]
        for dep in plugin.deployments.get_all_entities():
            row = [dep.deployment_uid, dep.name, dep.disabled, dep.created, dep.updated]
            agents = [x.agent_uid for x in plugin.deployment_agents.get_links_for_subject(dep.deployment_uid)]
            if verbose:
                row.append(agents)
            else:
                row.append(len(agents))
            table.append(row)

        table.sort(key=lambda x: x[1])
        fmt = kwargs.get('format')
        if fmt != 'json':
            headers = [report_utils.field_to_title(x) for x in headers]
        return report_utils.dump_report_data(table, headers, fmt=fmt, filename=kwargs.get('output'))


class PedmDeploymentAddCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='add', description='Add PEDM deployments')
        parser.add_argument('-f', '--force', dest='force', action='store_true',
                            help='do not prompt for confirmation')
        # parser.add_argument('--spiffe-cert', dest='spiffe', action='store',
        #                     help='File containing SPIFFE server certificate')
        parser.add_argument('name', help='Deployment name')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs):
        enterprise_data = context.enterprise
        assert enterprise_data is not None
        plugin = admin_plugin.get_pedm_plugin(context)
        deployment_name = kwargs.get('name')
        if not deployment_name:
            raise base.CommandError('Deployment name is required')
        force =kwargs.get('force') is True
        if not force:
            l_name = deployment_name.lower()
            has_name = any((True for x in plugin.deployments.get_all_entities() if x.name.lower() == l_name))
            if has_name:
                raise base.CommandError(f'Deployment "{deployment_name}" already exists.')

        enterprise_keys = enterprise_data['keys']
        ec_public = enterprise_keys['ecc_public_key']
        ec_public_key = utils.base64_url_decode(ec_public)
        #ec_public_key = crypto.unload_ec_public_key(enterprise_data.enterprise_info.ec_public_key)
        agent_info = pedm_shared.DeploymentAgentInformation(hash_key=plugin.agent_key, peer_public_key=ec_public_key)
        spiffe_cert: Optional[bytes] = None
        spiffe = kwargs.get('spiffe')
        if isinstance(spiffe, str):
            spiffe = os.path.expanduser(spiffe)
            if not os.path.isfile(spiffe):
                raise base.CommandError(f'File "{spiffe}" does not exist')
            _, ext = os.path.splitext(spiffe)
            with open(spiffe, 'rb') as f:
                if ext in ('.cer', '.der'):
                    cert = x509.load_der_x509_certificate(f.read())
                elif ext == '.pem':
                    cert = x509.load_pem_x509_certificate(f.read())
                else:
                    cert = x509.load_pem_x509_certificate(f.read())
                spiffe_cert = cert.public_bytes(serialization.Encoding.DER)
        add_rq = admin_types.AddDeployment(name=deployment_name, spiffe_cert=spiffe_cert, agent_info=agent_info)
        rs = plugin.modify_deployments(add_deployments=[add_rq])
        if len(rs.remove) > 0:
            status = rs.remove[0]
            if isinstance(status, admin_types.EntityStatus) and not status.success:
                raise base.CommandError(f'Failed to add deployment "{status.entity_uid}": {status.message}')


class PedmDeploymentUpdateCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='update', description='Update PEDM deployment')
        parser.add_argument('--disable', dest='disable', action='store', choices=['on', 'off'],
                            help='do not prompt for confirmation')
        # parser.add_argument('--spiffe-cert', dest='spiffe', action='store',
        #                     help='File containing SPIFFE server certificate')
        parser.add_argument('--name',action='store', help='Deployment name')
        parser.add_argument('deployment', metavar='DEPLOYMENT', help='Deployment name or UID')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs):
        plugin = admin_plugin.get_pedm_plugin(context)
        deployment = PedmUtils.resolve_single_deployment(plugin, kwargs.get('deployment'))
        name = kwargs.get('name')
        disable_choice = kwargs.get('disable')
        disabled: Optional[bool] = None
        if disable_choice is not None:
            disabled = True if disable_choice == 'on' else False

        spiffe_cert: Optional[bytes] = None
        spiffe = kwargs.get('spiffe')
        if isinstance(spiffe, str):
            spiffe = os.path.expanduser(spiffe)
            if not os.path.isfile(spiffe):
                raise base.CommandError(f'File "{spiffe}" does not exist')
            _, ext = os.path.splitext(spiffe)
            with open(spiffe, 'rb') as f:
                if ext in ('.cer', '.der'):
                    cert = x509.load_der_x509_certificate(f.read())
                elif ext == '.pem':
                    cert = x509.load_pem_x509_certificate(f.read())
                else:
                    cert = x509.load_pem_x509_certificate(f.read())
                spiffe_cert = cert.public_bytes(serialization.Encoding.DER)

        update_rq = admin_types.UpdateDeployment(
            deployment_uid=deployment.deployment_uid, name=name, disabled=disabled, spiffe_cert=spiffe_cert)
        rs = plugin.modify_deployments(update_deployments=[update_rq])
        if len(rs.remove) > 0:
            status = rs.remove[0]
            if isinstance(status, admin_types.EntityStatus) and not status.success:
                raise base.CommandError(f'Failed to update policy "{status.entity_uid}": {status.message}')


class PedmDeploymentDeleteCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='delete', description='Delete PEDM deployment')
        parser.add_argument('-f', '--force', dest='force', action='store_true',
                            help='do not prompt for confirmation')
        parser.add_argument('deployment', metavar='DEPLOYMENT', nargs='+',
                            help='Deployment name or UID')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs):
        plugin = admin_plugin.get_pedm_plugin(context)
        deployment_names = kwargs.get('deployment')
        if isinstance(deployment_names, str):
            deployment_names = [deployment_names]
        if not isinstance(deployment_names, list):
            raise base.CommandError(f'deployment argument is empty')

        deployments: List[str] = []
        for deployment_name in deployment_names:
            try:
                deployment = PedmUtils.resolve_single_deployment(plugin, deployment_name)
                deployment_name = deployment.name
                deployment_uid = deployment.deployment_uid
            except Exception as e:
                d = plugin.storage.deployments.get_entity(deployment_name) if deployment_name else None
                if d:
                    deployment_uid = d.deployment_uid
                else:
                    raise e
            deployments.append(deployment_uid)

        if len(deployments) == 0:
            raise base.CommandError('No deployments found')

        force = kwargs.get('force') is True
        if not force:
            answer = prompt_utils.user_choice(f'Do you want to delete {len(deployments)} deployment(s)?', 'yN')
            if answer.lower() not in {'y', 'yes'}:
                return

        rs = plugin.modify_deployments(remove_deployments=deployments)
        for status in rs.remove:
            if isinstance(status, admin_types.EntityStatus) and not status.success:
                raise base.CommandError(f'Failed to delete deployment "{status.entity_uid}": {status.message}')


class PedmDeploymentDownloadCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='download', description='Download PEDM deployment package')
        grp = parser.add_mutually_exclusive_group()
        grp.add_argument('--file', dest='file', action='store', help='File name')
        grp.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='Verbose output')
        parser.add_argument('deployment', metavar='DEPLOYMENT', help='Deployment name or UID')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> Optional[str]:
        plugin = admin_plugin.get_pedm_plugin(context)

        deployment = PedmUtils.resolve_single_deployment(plugin, kwargs.get('deployment'))
        host = next((host for host, server in constants.KEEPER_PUBLIC_HOSTS.items() if server == context.server), context.server)
        token = f'{host}:{deployment.deployment_uid}:{utils.base64_url_encode(deployment.private_key)}'
        filename = kwargs.get('file')
        if filename:
            with open(filename, 'wt') as f:
                f.write(token)
                return None

        if not kwargs.get('verbose'):
            return token

        path = ''
        windows = ''
        macos = ''
        linux = ''

        try:
            hostname = whoami.get_hostname(context.rest_context.server_base)
            for dc in constants.KEEPER_PUBLIC_HOSTS.values():
                if hostname.endswith(dc):
                    us = constants.KEEPER_PUBLIC_HOSTS['US']
                    hostname = hostname[:-len(us)] + us
                    break

            manifest_url = urlunparse(('https', hostname, '/pam/pedm/package-manifest.json', None, None, None))
            rs = requests.get(manifest_url)
            manifest = rs.json()
            core = manifest.get('Core')
            if isinstance(core, list) and len(core) > 0:
                latest = core[0]
                path = latest.get('Path')
                windows = latest.get('WindowsZip')
                macos = latest.get('MacOsZip')
                linux = latest.get('LinuxZip')
        except:
            pass

        table = [['', '']]
        if path:
            if windows:
                table.append(['Windows download URL', path + windows])
            if macos:
                table.append(['MacOS download URL', path + macos])
            if linux:
                table.append(['Linux download URL', path + linux])
            table.append(['', ''])
        table.append(['Deployment Token', token])
        base.dump_report_data(table, ['key', 'value'], no_header=True)
        return None


class PedmAgentCommand(base.GroupCommandNew):
    def __init__(self):
        super().__init__('Manage PEDM agents')
        self.register_command_new(PedmAgentListCommand(), 'list', 'l')
        self.register_command_new(PedmAgentEditCommand(), 'edit', 'e')
        self.register_command_new(PedmAgentDeleteCommand(), 'delete')
        self.register_command_new(PedmAgentCollectionCommand(), 'collection', 'c')
        self.default_verb = 'list'


class PedmAgentCollectionCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='list', parents=[base.report_output_parser],
                                         description='List PEDM agent resources')
        parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                            help='print verbose information')
        parser.add_argument('--type', dest='type', action='store', type=int,
                            help='collection type filter')
        parser.add_argument('agent', help='Agent UID')

        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> Any:
        plugin = admin_plugin.get_pedm_plugin(context)

        verbose = kwargs.get('verbose') is True
        collection_type: Optional[int] = kwargs.get('type')
        agent = PedmUtils.resolve_single_agent(plugin, kwargs.get('agent'))
        resource_uids = {x.collection_uid for x in plugin.storage.collection_links.get_links_for_object(agent.agent_uid)}
        collections = [plugin.collections.get_entity(x) or x for x in resource_uids]
        if isinstance(collection_type, int):
            collections = [x for x in collections if isinstance(x, admin_types.PedmCollection) and x.collection_type == collection_type]

        table: List[List[Any]] = []
        headers = ['collection_type']
        if verbose:
            headers.extend(['collection_uid', 'value'])
            for collection in collections:
                if isinstance(collection, admin_types.PedmCollection):
                    col_type_name = pedm_shared.collection_type_to_name(collection.collection_type)
                    col_type_name += f' ({col_type_name})'
                    collection_value = [f'{k}={v}' for k, v in collection.collection_data.items()]
                    row = [col_type_name, collection.collection_uid, collection_value]
                else:
                    row = ['', collection, '']
                table.append(row)
        else:
            headers.extend(['count'])
            r_map: Dict[int, int] = {}
            for collection in collections:
                if not isinstance(collection, admin_types.PedmCollection):
                    continue
                if collection.collection_type not in r_map:
                    r_map[collection.collection_type] = 0
                r_map[collection.collection_type] += 1
            for collection_type, cnt in r_map.items():
                col_type_name = pedm_shared.collection_type_to_name(collection_type)
                col_type_name += f' ({collection_type})'
                table.append([col_type_name, cnt])

        table.sort(key=lambda x: x[0])
        fmt = kwargs.get('format')
        if fmt != 'json':
            headers = [report_utils.field_to_title(x) for x in headers]
        return report_utils.dump_report_data(table, headers, fmt=fmt, filename=kwargs.get('output'))


class PedmAgentDeleteCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='update', description='Delete PEDM agents')
        parser.add_argument('--force', dest='force', action='store_true',
                            help='do not prompt for confirmation')
        parser.add_argument('agent', nargs='+', help='Agent UID(s)')

        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> Any:
        plugin = admin_plugin.get_pedm_plugin(context)
        agents = kwargs['agent']
        if isinstance(agents, str):
            agents = [agents]
        agent_uid_list: List[str] = []
        if isinstance(agents, list):
            for agent_name in agents:
                agent = PedmUtils.resolve_single_agent(plugin, agent_name)
                agent_uid_list.append(agent.agent_uid)

        if len(agent_uid_list) == 0:
            return

        statuses = plugin.modify_agents( remove_agents=agent_uid_list)
        if isinstance(statuses.remove, list):
            for status in statuses.remove:
                if isinstance(status, admin_types.EntityStatus) and not status.success:
                    utils.get_logger().warning(f'Failed to remove agent "{status.entity_uid}": {status.message}')


class PedmAgentEditCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='update', description='Update PEDM agents')
        parser.add_argument('--enable', dest='enable', action='store', choices=['on', 'off'],
                                   help='Enables or disables agents')
        parser.add_argument('--deployment', dest='deployment', action='store',
                                   help='Moves agent to deployment')
        parser.add_argument('agent', nargs='+', help='Agent UID(s)')

        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> Any:
        plugin = admin_plugin.get_pedm_plugin(context)

        deployment_uid = kwargs.get('deployment')
        if deployment_uid:
            deployment = plugin.deployments.get_entity(deployment_uid)
            if not deployment:
                raise base.CommandError(f'Deployment "{deployment_uid}" does not exist')
        else:
            deployment_uid = None

        disabled: Optional[bool] = None
        enable = kwargs.get('enable')
        if isinstance(enable, str):
            if enable.lower() == 'on':
                disabled = False
            elif enable.lower() == 'off':
                disabled = True
            else:
                raise base.CommandError(f'"enable" argument must be "on" or "off"')

        update_agents: List[admin_types.UpdateAgent] = []
        agents = kwargs['agent']
        if isinstance(agents, str):
            agents = [agents]
        if isinstance(agents, list):
            for a in agents:
                agent = plugin.agents.get_entity(a)
                if agent is None:
                    raise base.CommandError(f'Agent "{a}" does not exist')
                update_agents.append(admin_types.UpdateAgent(
                    agent_uid=agent.agent_uid,
                    deployment_uid=deployment_uid,
                    disabled=disabled,
                ))
        if len(update_agents) > 0:
            statuses = plugin.modify_agents(update_agents=update_agents)
            if isinstance(statuses.update, list):
                for status in statuses.update:
                    if isinstance(status, admin_types.EntityStatus) and not status.success:
                        utils.get_logger().warning(f'Failed to update agent "{status.entity_uid}": {status.message}')


class PedmAgentListCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='list', description='List PEDM agents',
                                         parents=[base.report_output_parser])
        parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                            help='print verbose information')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> Any:
        plugin = admin_plugin.get_pedm_plugin(context)

        verbose = kwargs.get('verbose') is True
        table = []
        headers = ['agent_uid', 'machine_name', 'deployment', 'disabled', 'created']
        active_agents: Set[str] = set()
        if verbose:
            headers.extend(('active', 'properties'))
            rq = pedm_pb2.PolicyAgentRequest()
            rq.summaryOnly = False
            rs = api.execute_router(context, "pedm/get_policy_agents", rq, rs_type=pedm_pb2.PolicyAgentResponse)
            assert rs is not None
            active_agents.update((utils.base64_url_encode(x) for x in rs.agentUid))

        for agent in plugin.agents.get_all_entities():
            deployment = plugin.deployments.get_entity(agent.deployment_uid)
            deployment_name = deployment.name if deployment else agent.deployment_uid
            time_created = datetime.datetime.fromtimestamp(int(agent.created // 1000)) if agent.created else None
            machine_name = ''
            if isinstance(agent.properties, dict):
                machine_name = agent.properties.get('MachineName') or ''
            row: List[Any] = [agent.agent_uid, machine_name, deployment_name, agent.disabled, time_created]
            if verbose:
                row.append(agent.agent_uid in active_agents)
                props: Optional[List[str]] = None
                if agent.properties:
                    props = [f'{k}={v}' for k, v in agent.properties.items()]
                    props.sort()
                row.append(props)

            table.append(row)

        table.sort(key=lambda x: x[2])
        fmt = kwargs.get('format')
        if fmt != 'json':
            headers = [report_utils.field_to_title(x) for x in headers]
        return report_utils.dump_report_data(table, headers, fmt=fmt, filename=kwargs.get('output'))


class PedmPolicyCommand(base.GroupCommandNew):
    def __init__(self):
        super().__init__('Manage PEDM policies')
        self.register_command_new(PedmPolicyListCommand(), 'list', 'l')
        self.register_command_new(PedmPolicyAddCommand(), 'add', 'a')
        self.register_command_new(PedmPolicyEditCommand(), 'edit', 'e')
        self.register_command_new(PedmPolicyViewCommand(), 'view', 'v')
        self.register_command_new(PedmPolicyAgentsCommand(), 'agents')
        self.register_command_new(PedmPolicyAssignCommand(), 'assign')
        self.register_command_new(PedmPolicyDeleteCommand(), 'delete')
        self.default_verb = 'list'


class PedmPolicyMixin:
    ALL_FILTERS: List[str] = ['USER', 'MACHINE', 'APP', 'TIME', 'DATE', 'DAY']
    ALL_CONTROLS: List[str] = ['ALLOW', 'DENY', 'NOTIFY', 'MFA', 'JUSTIFY', 'APPROVAL', 'AUDIT', 'RECORD']

    policy_filter = argparse.ArgumentParser(add_help=False)
    policy_filter.add_argument('--user-filter', dest='user_filter', action='append',
                        help='Policy user filter. User collection UID or *')
    policy_filter.add_argument('--machine-filter', dest='machine_filter', action='append',
                        help='Policy machine filter. Machine collection UID ')
    policy_filter.add_argument('--app-filter', dest='app_filter', action='append',
                        help='Policy application filter. Application collection UID')
    policy_filter.add_argument('--date-filter', dest='date_filter', action='append',
                        help='Policy date filter. Date range in ISO format. YYYY-MM-DD:YYYY-MM-DD')
    policy_filter.add_argument('--time-filter', dest='time_filter', action='append',
                        help='Policy time filter. Time. 24 hours format: HH:MM-HH:MM')
    policy_filter.add_argument('--day-filter', dest='day_filter', action='append',
                        help='Policy day filter. Day of Week')
    policy_filter.add_argument('--risk-level', dest='risk_level', type=int, help='Policy risk level')

    @staticmethod
    def resolve_collections(plugin: admin_plugin.PedmPlugin, col_types: List[int], col_values: List[str]) -> List[str]:
        result: List[str] = []
        if not col_values:
            return result

        collection_lookup: Dict[str, Union[str, List[str]]] = {}
        for c in plugin.collections.get_all_entities():
            if c.collection_type not in col_types: continue
            collection_lookup[c.collection_uid] = c.collection_uid
            if c.collection_type >= 100:
                collection_name: Optional[str] = c.collection_data.get('Name')
                if not collection_name:
                    continue
                collection_name = collection_name.lower()
                cv = collection_lookup.get(collection_name)
                if not cv:
                    cv = c.collection_uid
                elif isinstance(cv, str):
                    cv = [cv, c.collection_uid]
                elif isinstance(cv, list):
                    cv.append(c.collection_uid)
                else:
                    continue
                collection_lookup[collection_name] = cv

        for col_value in col_values:
            if col_value == '*':
                result.append(col_value)
            else:
                cv = collection_lookup[col_value]
                if not cv:
                    cv = collection_lookup[col_value.lower()]
                if not cv:
                    raise base.CommandError(f'collection value "{col_value}" cannot be resolved')
                if isinstance(cv, str):
                    result.append(cv)
                else:
                    raise base.CommandError(f'collection value "{col_value}" is not unique. Use collection UID')

        return result

    @staticmethod
    def to_time(v: str) -> Optional[str]:
        if not v:
            return None

        try:
            tc = [int(x) for x in v.split(':')]
            while len(tc) < 3:
                tc.append(0)
            if tc[0] >= 24:
                raise base.CommandError(f'time value "{v}" is not valid. Hours: 0 - 23')
            if tc[1] >= 60:
                raise base.CommandError(f'time value "{v}" is not valid. Minutes: 0 - 59')
            if tc[2] >= 60:
                raise base.CommandError(f'time value "{v}" is not valid. Seconds: 0 - 59')

            return ':'.join((f'{x:02d}' for x in tc))
        except Exception as e:
            raise base.CommandError(f'time value "{v}" is not valid.')

    @staticmethod
    def from_time(v: Any) -> Optional[str]:
        if not isinstance(v, str):
            return None
        try:
            tc = [int(x) for x in v.split(':')]
            tc = tc[:3]
            if tc[2] == 0:
                tc = tc[:2]
            return ':'.join((f'{x:02d}' for x in tc))
        except Exception:
            pass

    @staticmethod
    def parse_times(policy_times: Optional[List[Dict[str, Any]]]) -> Optional[List[str]]:
        if not isinstance(policy_times, list):
            return None

        result: List[str] = []
        for policy_time in policy_times:
            start_time = PedmPolicyMixin.from_time(policy_time.get('StartTime')) or ''
            end_time = PedmPolicyMixin.from_time(policy_time.get('EndTime')) or ''
            if start_time or end_time:
                result.append(f'{start_time}-{end_time}')
        return result

    @staticmethod
    def to_date(v: str) -> Optional[str]:
        if not v:
            return None
        try:
            date_value = datetime.datetime.fromisoformat(v).date()
            return date_value.isoformat()
        except Exception as e:
            raise base.CommandError(f'date value "{v}" is not valid.')

    @staticmethod
    def resolve_dates(d_values: List[str]) -> List[Dict[str, str]]:
        # { "StartDate": "2025-01-01",  "EndDate": "2025-01-25" }
        result: List[Dict[str, str]] = []
        if not d_values:
            return result
        for d_value in d_values:
            comp: List[Any] = d_value.split(':')
            if 1 <= len(comp) <= 2:
                dat: Dict[str, str] = {}
                comp = [PedmPolicyAddCommand.to_date(x) for x in comp]
                if comp[0]:
                    dat['StartDate'] = comp[0]
                if len(comp) == 2 and comp[1]:
                    dat['EndDate'] = comp[1]
                result.append(dat)
            else:
                raise base.CommandError(f'date range "{d_value}" is not valid.')

        return result

    @staticmethod
    def resolve_times(t_values: List[str]) -> List[Dict[str, str]]:
        #   { "StartTime" : "09:00:00", "EndTime" : "18:00:00" }
        result: List[Dict[str, str]] = []
        if not t_values:
            return result
        for t_value in t_values:
            comp: List[Any] = t_value.split('-')
            if 1 <= len(comp) <= 2:
                tim: Dict[str, str] = {}
                comp = [PedmPolicyAddCommand.to_time(x) for x in comp]
                if comp[0]:
                    tim['StartTime'] = comp[0]
                if len(comp) == 2 and comp[1]:
                    tim['EndTime'] = comp[1]
                result.append(tim)
            else:
                raise base.CommandError(f'time range "{t_value}" is not valid.')

        return result

    DAY_LOOKUP: Optional[Dict[str, int]] = None
    @staticmethod
    def get_day_lookup() -> Dict[str, int]:
        if PedmPolicyMixin.DAY_LOOKUP is None:
            PedmPolicyMixin.DAY_LOOKUP = {}
            for day_no, day_name in enumerate(calendar.day_name):
                day_no += 1
                if day_no > 6:
                    day_no -= 7
                PedmPolicyMixin.DAY_LOOKUP[day_name.lower()] = day_no
            for day_no, day_name in enumerate(calendar.day_abbr):
                day_no += 1
                if day_no > 6:
                    day_no -= 7
                PedmPolicyMixin.DAY_LOOKUP[day_name.lower()] = day_no
        return PedmPolicyMixin.DAY_LOOKUP

    @staticmethod
    def resolve_days(d_values: List[str]) -> List[int]:
        # integer in American convention
        result: List[int] = []
        if not d_values:
            return result

        day_lookup = PedmPolicyMixin.get_day_lookup()
        weekday: Optional[int]
        for d_value in d_values:
            if d_value.isnumeric():
                weekday = int(d_value)
                if 6 < weekday < 0:
                    weekday = None
            else:
                weekday = day_lookup.get(d_value.lower())
            if weekday is None:
                raise base.CommandError(f'day value "{d_value}" is not valid.')
            result.append(weekday)
        return result

    @staticmethod
    def get_policy_controls(policy_type_name: str, **kwargs) -> Optional[List[str]]:
        p_controls: Optional[Union[str, List[str]]] = kwargs.get('control')
        if not p_controls:
            return None

        allowed_controls: Set[str] = set()
        if policy_type_name == 'PrivilegeElevation':
            allowed_controls.update(('audit', 'notify', 'mfa', 'justify', 'approval'))
        elif policy_type_name == 'Access':
            allowed_controls.update(('audit', 'notify', 'allow', 'deny'))
        elif policy_type_name == 'CommandLine':
            allowed_controls.update(('audit', 'notify', 'allow', 'deny'))

        controls: List[str] = []
        if isinstance(p_controls, str):
            controls = [p_controls]

        wrong_controls = set(p_controls) - allowed_controls
        if len(wrong_controls) > 0:
            raise base.CommandError(f'"Control(s): {(", ".join(wrong_controls))}" are not valid for {policy_type_name} policy type')

        p_c = {x.upper() for x in p_controls}
        for c in PedmPolicyMixin.ALL_CONTROLS:
            if c in p_c:
                p_c.remove(c)
                controls.append(c)
        if len(p_c) > 0:
            raise base.CommandError(f'"control: {", ".join(p_c)}" is not supported')
        return controls

    @staticmethod
    def get_policy_filter(plugin: admin_plugin.PedmPlugin, **kwargs) -> Dict[str, Any]:
        policy_filter: Dict[str, Any] = {}
        for f in PedmPolicyMixin.ALL_FILTERS:
            arg_name = f'{f.lower()}_filter'
            p_filter: Any = kwargs.get(arg_name)
            if not p_filter: continue
            if isinstance(p_filter, str):
                p_filter = [p_filter]

            if f == 'USER':
                filter_name = 'UserCheck'
            elif f == 'MACHINE':
                filter_name = 'MachineCheck'
            elif f == 'APP':
                filter_name = 'ApplicationCheck'
            elif f == 'DATE':
                filter_name = 'DateCheck'
            elif f == 'TIME':
                filter_name = 'TimeCheck'
            elif f == 'DAY':
                filter_name = 'DayCheck'
            else:
                continue
            if '*' in p_filter:
                policy_filter[filter_name] = ['*']
            else:
                if f == 'USER':
                    policy_filter[filter_name] = PedmPolicyAddCommand.resolve_collections(plugin, [3, 6, 103], p_filter)
                elif f == 'MACHINE':
                    policy_filter[filter_name] = PedmPolicyAddCommand.resolve_collections(plugin, [1, 101], p_filter)
                elif f == 'APP':
                    policy_filter[filter_name] = PedmPolicyAddCommand.resolve_collections(plugin, [2, 102], p_filter)
                elif f == 'DATE':
                    policy_filter[filter_name] = PedmPolicyAddCommand.resolve_dates(p_filter)
                elif f == 'TIME':
                    policy_filter[filter_name] = PedmPolicyAddCommand.resolve_times(p_filter)
                elif f == 'DAY':
                    policy_filter[filter_name] = PedmPolicyAddCommand.resolve_days(p_filter)
        risk_level = kwargs.get('risk_level')
        if isinstance(risk_level, int):
            if risk_level < 0 or risk_level > 100:
                raise base.CommandError(f'risk level "{risk_level}" is not valid: 0-100')
            policy_filter['RiskLevel'] = risk_level
        return policy_filter


class PedmPolicyListCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='list', description='List PEDM policies',
                                         parents=[base.report_output_parser])
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> Any:
        plugin = admin_plugin.get_pedm_plugin(context)
        table: List[List[Any]] = []
        all_agents = utils.base64_url_encode(plugin.all_agents)
        headers = ['policy_uid', 'policy_name', 'policy_type', 'status', 'controls', 'users', 'machines', 'applications', 'collections']
        for policy in plugin.policies.get_all_entities():
            data = policy.data or {}
            actions = data.get('Actions') or {}
            on_success = actions.get('OnSuccess') or {}
            controls = on_success.get('Controls') or ''

            collections = [x.collection_uid for x in plugin.storage.collection_links.get_links_for_object(policy.policy_uid)]
            collections = ['*' if x == all_agents else x for x in collections]
            collections.sort()

            status = data.get('Status')
            if policy.disabled:
                status = 'off'
            table.append([policy.policy_uid, data.get('PolicyName'), data.get('PolicyType'), status,
                          controls, data.get('UserCheck'), data.get('MachineCheck'), data.get('ApplicationCheck'),
                          collections])

        fmt = kwargs.get('format')
        if fmt != 'json':
            headers = [report_utils.field_to_title(x) for x in headers]
        return report_utils.dump_report_data(table, headers, fmt=fmt, filename=kwargs.get('output'), sort_by=1)


class PedmPolicyAddCommand(base.ArgparseCommand, PedmPolicyMixin):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='add', description='Add PEDM policy', parents=[PedmPolicyMixin.policy_filter])
        parser.add_argument('--policy-type', dest='policy_type', action='store', default='elevation',
                            choices=['elevation', 'file_access', 'command', 'least_privilege'],
                            help='Policy type')
        parser.add_argument('--policy-name', dest='policy_name', action='store',
                            help='Policy name')
        parser.add_argument('--control', dest='control', action='append',
                            choices=['allow', 'deny', 'audit', 'notify', 'mfa', 'justify', 'approval'],
                            help='Policy controls')
        parser.add_argument('--status', dest='status', action='store',
                            choices=['enforce', 'monitor', 'monitor_and_notify'],
                            help='Policy Status')
        parser.add_argument('--enable', dest='enable', action='store', choices=['on', 'off'],
                            help='Enables or disables policy')

        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> None:
        plugin = admin_plugin.get_pedm_plugin(context)

        p_type = kwargs.get('policy_type')
        if p_type == 'elevation':
            policy_type = 'PrivilegeElevation'
        elif p_type == 'file_access':
            policy_type = 'FileAccess'
        elif p_type == 'command':
            policy_type = 'CommandLine'
        elif p_type == 'least_privilege':
            policy_type = 'LeastPrivilege'
        else:
            raise base.CommandError(f'"policy-type: {p_type}" is not supported')

        policy_uid = utils.generate_uid()
        controls = PedmPolicyMixin.get_policy_controls(policy_type, **kwargs)

        policy_data: Dict[str, Any] = {
            'PolicyName': kwargs.get('policy_name') or '',
            'PolicyType': policy_type,
            'PolicyId': policy_uid,
            'Status': 'off',
            'Actions': {
                'OnSuccess': {'Controls': controls or []},
                'OnFailure': {'Command': ''}
            },
            "NotificationMessage": "A policy has been set to monitor mode.  When this policy is enabled, [mfa, justification, request] will be required to run this process as an administrator.",
            "NotificationRequiresAcknowledge": False,
            "RiskLevel": 50,
            'Operator': 'And',
            'Rules': [
                {
                    'RuleName': 'UserCheck',
                    'ErrorMessage': 'This user is not included in this policy',
                    'RuleExpressionType': 'BuiltInAction',
                    'Expression': 'CheckUser()'
                },
                {
                    'RuleName': 'MachineCheck',
                    'ErrorMessage': 'This Machine is not included in this policy',
                    'RuleExpressionType': 'BuiltInAction',
                    'Expression': 'CheckMachine()'
                },
                {
                    'RuleName': 'ApplicationCheck',
                    'ErrorMessage': 'This application is not included in this policy',
                    'RuleExpressionType': 'BuiltInAction',
                    'Expression': 'CheckFile(false)'
                },
                {
                    "RuleName": "DateCheck",
                    "ErrorMessage": "Current date is not covered by this policy",
                    "RuleExpressionType": "BuiltInAction",
                    "Expression": "CheckDate()"
                },
                {
                    'RuleName': 'TimeCheck',
                    'ErrorMessage': 'Current time is not covered by this policy',
                    'RuleExpressionType': 'BuiltInAction',
                    'Expression': 'CheckTime()'
                },
                {
                    'RuleName': 'DayCheck',
                    'ErrorMessage': 'Today is not included in this policy',
                    'RuleExpressionType': 'BuiltInAction',
                    'Expression': 'CheckDay()'
                }
            ]
        }
        policy_filter = PedmPolicyMixin.get_policy_filter(plugin, **kwargs)
        if policy_filter:
            policy_data.update(policy_filter)

        for filter_name in ('UserCheck', 'MachineCheck', 'ApplicationCheck', 'DateCheck', 'TimeCheck', 'DayCheck'):
            f = policy_filter.get(filter_name)
            if f is None:
                policy_filter[filter_name] = ['*']

        arg_status = kwargs.get('status')
        if isinstance(arg_status, str):
            policy_data['Status'] = arg_status
        else:
            policy_data['Status'] = 'enforce'

        disabled: bool = False
        arg_enable = kwargs.get('enable')
        if isinstance(arg_enable, str):
            disabled = True if arg_enable == 'off' else False

        policy_key = utils.generate_aes_key()
        add_policy = admin_types.PedmPolicy(
            policy_uid=policy_uid, policy_key=policy_key, data=policy_data, admin_data={}, disabled=disabled)
        rs = plugin.modify_policies(add_policies=[add_policy])
        if len(rs.remove) > 0:
            status = rs.remove[0]
            if isinstance(status, admin_types.EntityStatus) and not status.success:
                raise base.CommandError(f'Failed to add policy "{status.entity_uid}": {status.message}')


class PedmPolicyEditCommand(base.ArgparseCommand, PedmPolicyMixin):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='edit', description='Edit PEDM policy', parents=[PedmPolicyMixin.policy_filter])
        parser.add_argument('policy', help='Policy UID')
        parser.add_argument('--policy-name', dest='policy_name', action='store',
                            help='Policy name')
        parser.add_argument('--control', dest='control', action='append',
                            choices=['allow', 'deny', 'audit', 'notify', 'mfa', 'justify', 'approval'],
                            help='Policy controls')
        parser.add_argument('--status', dest='status', action='store',
                            choices=['enforce', 'monitor', 'monitor_and_notify'],
                            help='Policy Status')
        parser.add_argument('--enable', dest='enable', action='store', choices=['on', 'off'],
                            help='Enables or disables policy')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> None:
        plugin = admin_plugin.get_pedm_plugin(context)

        policy = PedmUtils.resolve_single_policy(plugin, kwargs.get('policy'))

        policy_data = copy.deepcopy(policy.data or {})
        policy_type = policy_data.get('PolicyType') or 'Unknown'
        controls = PedmPolicyMixin.get_policy_controls(policy_type, **kwargs)
        if isinstance(controls, list):
            actions = policy_data.get('Actions')
            if not isinstance(actions, dict):
                actions = {}
                policy_data['Actions'] = actions
            on_success = actions.get('OnSuccess')
            if not isinstance(on_success, dict):
                on_success = {}
            on_success['Controls'] = controls
            policy_data['OnSuccess'] = on_success

        policy_name = kwargs.get('policy_name')
        if policy_name:
            policy_data['PolicyName'] = policy_name
        policy_filter = PedmPolicyMixin.get_policy_filter(plugin, **kwargs)
        if policy_filter:
            policy_data.update(policy_filter)

        arg_status = kwargs.get('status')
        if isinstance(arg_status, str):
            policy_data['Status'] = arg_status

        disabled: Optional[bool] = None
        arg_enable = kwargs.get('enable')
        if isinstance(arg_enable, str):
            disabled = True if arg_enable == 'off' else False

        pu = admin_types.PedmUpdatePolicy(policy_uid=policy.policy_uid, data=policy_data, disabled=disabled)

        rs = plugin.modify_policies(update_policies=[pu])
        if len(rs.update) > 0:
            status = rs.update[0]
            if isinstance(status, admin_types.EntityStatus) and not status.success:
                raise base.CommandError(f'Failed to update policy "{status.entity_uid}": {status.message}')


class PedmPolicyViewCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='view', parents=[base.json_output_parser], description='View PEDM policy')
        parser.add_argument('policy', help='Policy UID or name')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> Any:
        plugin = admin_plugin.get_pedm_plugin(context)

        policy = PedmUtils.resolve_single_policy(plugin, kwargs.get('policy'))

        body = json.dumps(policy.data, indent=4)
        filename = kwargs.get('output')
        if kwargs.get('format') == 'json' and filename:
            with open(filename, 'w') as f:
                f.write(body)
        else:
            return body


class PedmPolicyDeleteCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='delete', description='Delete PEDM policy')
        parser.add_argument('policy', type=str, nargs='+', help='Policy UID or name')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> None:
        plugin = admin_plugin.get_pedm_plugin(context)

        policies = PedmUtils.resolve_existing_policies(plugin, kwargs.get('policy'))
        to_delete = [x.policy_uid for x in policies]

        rs = plugin.modify_policies(remove_policies=to_delete)
        if len(rs.remove) > 0:
            status = rs.remove[0]
            if isinstance(status, admin_types.EntityStatus) and not status.success:
                raise base.CommandError(f'Failed to delete policy "{status.entity_uid}": {status.message}')


class PedmPolicyAgentsCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='agent', description='Show agents for policies')
        parser.add_argument('policy', type=str, nargs='+', help='Policy UID or name')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> None:
        plugin = admin_plugin.get_pedm_plugin(context)

        policy_args = kwargs.get('policy')
        if not isinstance(policy_args, list):
            policy_args = [policy_args]
        policies = PedmUtils.resolve_existing_policies(plugin, policy_args)
        if len(policies) == 0:
            policy_list = ', '.join(policy_args)
            raise base.CommandError(f'Policy "{policy_list}" not found')
        policy_uids = [utils.base64_url_decode(x.policy_uid) for x in policies]
        rq = pedm_pb2.PolicyAgentRequest()
        rq.policyUid.extend(policy_uids)
        rq.summaryOnly = False
        rs = api.execute_router(context, "pedm/get_policy_agents", rq, rs_type=pedm_pb2.PolicyAgentResponse)
        assert rs is not None

        table = []
        headers = ['key', 'uid', 'name', 'status']
        for p in policies:
            data = p.data or {}
            status = data.get('Status')
            if p.disabled:
                status = 'off'
            table.append(['Policy', p.policy_uid, data.get('PolicyName'), status])
        for a in rs.agentUid:
            agent_uid = utils.base64_url_encode(a)
            row = ['Agent', agent_uid]
            agent = plugin.agents.get_entity(agent_uid)
            machine_name = ''
            status = ''
            if agent:
                if isinstance(agent.properties, dict):
                    machine_name = agent.properties.get('MachineName') or ''
                status = 'off' if agent.disabled else 'on'
            row.append(machine_name)
            row.append(status)
            table.append(row)

        return report_utils.dump_report_data(table, headers, group_by=0)


class PedmPolicyAssignCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='assign', description='Assign collections to policy')
        parser.add_argument('-c', '--collection', action='append', help='Collection UID')
        parser.add_argument('policy', type=str, nargs='+', help='Policy UID or name')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> None:
        plugin = admin_plugin.get_pedm_plugin(context)

        policies = PedmUtils.resolve_existing_policies(plugin, kwargs.get('policy'))
        policy_uids = [utils.base64_url_decode(x.policy_uid) for x in policies]
        collections = kwargs.get('collection')
        collection_uids: List[bytes] = []
        if isinstance(collections, list):
            for c in collections:
                if c in ['*', 'all']:
                    collection_uids.append(plugin.all_agents)
                elif c:
                    collection_uid = utils.base64_url_decode(c)
                    if len(collection_uid) == 16:
                        collection_uids.append(collection_uid)
                    else:
                        utils.get_logger().info('Invalid collection UID: %s. Skipped', c)

        if len(policy_uids) == 0:
            raise base.CommandError('Nothing to do')

        statuses = plugin.assign_policy_collections(policy_uids, collection_uids)
        for status in statuses.add:
            if not status.success:
                raise base.CommandError(f'Failed to add to policy: {status.message}')
        for status in statuses.remove:
            if not status.success:
                raise base.CommandError(f'Failed to remove from policy: {status.message}')


class PedmCollectionCommand(base.GroupCommandNew):
    def __init__(self):
        super().__init__('Manage PEDM collections')
        self.register_command_new(PedmCollectionListCommand(), 'list', 'l')
        self.register_command_new(PedmCollectionViewCommand(), 'view', 'v')
        self.register_command_new(PedmCollectionAddCommand(), 'add', 'a')
        self.register_command_new(PedmCollectionUpdateCommand(), 'update', 'u')
        self.register_command_new(PedmCollectionDeleteCommand(), 'delete')
        self.register_command_new(PedmCollectionConnectCommand(), 'connect')
        self.register_command_new(PedmCollectionDisconnectCommand(), 'disconnect')
        self.register_command_new(PedmCollectionWipeOutCommand(), 'wipe-out')
        self.default_verb = 'list'


class PedmCollectionWipeOutCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='wipe-out', description='Wipe out PEDM collections')
        parser.add_argument('--type', dest='type', action='store', type=int,
                            help='collection type')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> None:
        plugin = admin_plugin.get_pedm_plugin(context)

        collection_type = kwargs.get('type')
        if isinstance(collection_type, int):
            collection_type = [collection_type]
        collections: List[str] = []
        for coll in plugin.storage.collections.get_all_entities():
            if collection_type and coll.collection_type not in collection_type:
                continue
            collections.append(coll.collection_uid)

        plugin.modify_collections(remove_collections=collections)


class PedmCollectionAddCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='add', description='Creates PEDM collections')
        parser.add_argument('--type', dest='type', action='store', type=int,
                            help='collection type')
        parser.add_argument('collection', type=str, nargs='+',  help='Collection name')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> None:
        plugin = admin_plugin.get_pedm_plugin(context)

        collection: Any = kwargs.get('collection')
        collection_type = kwargs.get('type')
        if not collection_type:
            raise base.CommandError('Collection type is required')
        # if collection_type < 100:
        #     raise base.CommandError('Only collections with type greater than 100 are supported')

        if isinstance(collection, str):
            collection = [collection]

        collections: Dict[str, admin_types.CollectionData] = {}
        for c in collection:
            collection_uid = utils.generate_uid()
            collection_data = {
                'Name': c,
                'IsCustom': True
            }
            collections[collection_uid] = admin_types.CollectionData(
                collection_uid=collection_uid, collection_type=collection_type,
                collection_data=json.dumps(collection_data))

        status = plugin.modify_collections(add_collections=collections.values())
        if len(status.add) > 0:
            for st in status.add:
                if isinstance(st, admin_types.EntityStatus) and not st.success:
                    raise base.CommandError(f'Failed to add collection "{st.entity_uid}": {st.message}')


class PedmCollectionUpdateCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='update', description='Update PEDM collection')
        parser.add_argument('--type', dest='type', action='store', type=int,
                            help='collection type (optional)')
        parser.add_argument('--name', dest='name', action='store', required=True,
                            help='Collection name')
        parser.add_argument('collection', help='Collection')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> None:
        plugin = admin_plugin.get_pedm_plugin(context)

        collection = kwargs.get('collection')
        collection_type = kwargs.get('type')
        collection_name = kwargs.get('name')
        if not collection_name:
            raise base.CommandError('Collection name is required')

        existing_collections = PedmUtils.resolve_existing_collections(plugin, collection, collection_type=collection_type)
        if len(existing_collections) > 0:
            if len(existing_collections) > 1:
                raise base.CommandError(f'Multiple collections found for collection "{collection}". Use Collection UID.')
            collections: admin_types.CollectionData
            coll = existing_collections[0]
            collection_info = coll.collection_data
            collection_info['Name'] = collection_name
            collection_data = admin_types.CollectionData(
                collection_uid=coll.collection_uid, collection_type=coll.collection_type,
                collection_data=json.dumps(collection_info))

            status = plugin.modify_collections(update_collections=[collection_data])
            if len(status.update) > 0:
                for st in status.update:
                    if isinstance(st, admin_types.EntityStatus) and not st.success:
                        raise base.CommandError(f'Failed to update collection "{st.entity_uid}": {st.message}')


class PedmCollectionDeleteCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='delete', description='Delete PEDM collections')
        parser.add_argument('-f', '--force', dest='force', action='store_true',
                            help='do not prompt for confirmation')
        parser.add_argument('collection', nargs='+', help='Collection or @orphan_resource')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> None:
        plugin = admin_plugin.get_pedm_plugin(context)

        collection = kwargs.get('collection')
        if not collection:
            raise base.CommandError('Collection is required')

        if isinstance(collection, str):
            collection = [collection]
        pseudo_collections = {x for x in collection if x in ('@orphan_resource')}
        collection = [x for x in collection if x not in pseudo_collections]
        force = kwargs.get('force') is True
        existing_collections = PedmUtils.resolve_existing_collections(plugin, collection, ignore_missing=True)
        unique_collections = set((x.collection_uid for x in existing_collections))
        if force:
            for collection_name in collection:
                if collection_name not in unique_collections:
                    try:
                        uid = utils.base64_url_decode(collection_name)
                        if len(uid) == 16:
                            unique_collections.add(collection_name)
                    except:
                        pass

        if '@orphan_resource' in pseudo_collections:
            unique_collections.update(PedmUtils.get_orphan_resources(plugin))

        if len(unique_collections) == 0:
            utils.get_logger().info('No collections found')
            return

        if not force:
            answer = prompt_utils.user_choice(f'Do you want to remove {len(unique_collections)} collection(s)?', 'yN', default='n')
            if answer.lower() not in ('y', 'yes'):
                return

        status = plugin.modify_collections(remove_collections=unique_collections)
        if len(status.remove) > 0:
            for st in status.remove:
                if isinstance(st, admin_types.EntityStatus) and not st.success:
                    raise base.CommandError(f'Failed to remove collection "{st.entity_uid}": {st.message}')


class PedmCollectionConnectCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='link', description='Link values to PEDM collection')
        parser.add_argument('--collection', '-c', dest='collection', action='store',
                            help='Parent collection UID or name')
        parser.add_argument('--link-type', dest='link_type', action='store', required=True,
                            choices=['agent', 'policy', 'collection'], help='collection type filter')
        parser.add_argument('links', nargs='+', help='Link UIDs or names')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> None:
        plugin = admin_plugin.get_pedm_plugin(context)

        col_name = kwargs.get('collection')
        collections = PedmUtils.resolve_existing_collections(plugin, [col_name])
        if len(collections) != 1:
            raise base.CommandError(f'Could not resolve a single collection: {col_name}')
        collection = collections[0]
        link_type = kwargs.get('link_type')
        link_names: Any = kwargs.get('links')
        links: List[str] = []
        collection_link_type: int
        if link_type == 'collection':
            coll_links = PedmUtils.resolve_existing_collections(plugin, link_names)
            links.extend((x.collection_uid for x in coll_links))
            collection_link_type = pedm_pb2.CLT_COLLECTION
        elif link_type == 'agent':
            for agent_name in link_names:
                agent = PedmUtils.resolve_single_agent(plugin, agent_name)
                links.append(agent.agent_uid)
            collection_link_type = pedm_pb2.CLT_AGENT
        elif link_type == 'policy':
            pol_links = PedmUtils.resolve_existing_policies(plugin, link_names)
            links.extend((x.policy_uid for x in pol_links))
            collection_link_type = pedm_pb2.CLT_POLICY
        else:
            raise base.CommandError(f'Unknown link type: {link_type}')

        to_add = [admin_types.CollectionLink(
            collection_uid=collection.collection_uid, link_uid=x, link_type=collection_link_type) for x in links]

        status = plugin.set_collection_links(set_links=to_add)
        if len(status.add) > 0:
            for st in status.add:
                if isinstance(st, admin_types.LinkStatus) and not st.success:
                    raise base.CommandError(f'Failed to set collection link "{st.object_uid}": {st.message}')


class PedmCollectionDisconnectCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='unlink', description='Unlink values from PEDM collections')
        parser.add_argument('--collection', '-c', dest='collection', action='store',
                            help='Parent collection UID or name')
        parser.add_argument('-f', '--force', dest='force', action='store_true',
                            help='do not prompt for confirmation')
        parser.add_argument('links', nargs='+', help='UIDs to unlink')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> None:
        plugin = admin_plugin.get_pedm_plugin(context)

        col_name = kwargs.get('collection')
        collections = PedmUtils.resolve_existing_collections(plugin, [col_name])
        if len(collections) != 1:
            raise base.CommandError(f'Could not resolve a single collection: {col_name}')
        collection = collections[0]

        existing_links= list(x for x in plugin.storage.collection_links.get_links_for_subject(collection.collection_uid))
        links: Any = kwargs.get('links')
        to_unlink: Set[str] = set(links)

        to_remove: List[admin_types.CollectionLink] = []
        for link in existing_links:
            link_uid = link.link_uid
            if link_uid in to_unlink:
                to_remove.append(admin_types.CollectionLink(
                    collection_uid=collection.collection_uid,
                    link_uid=link_uid,
                    link_type=link.link_type)     # type: ignore
                )
                to_unlink.remove(link_uid)

        if len(to_unlink) > 0:
            utils.get_logger().info(f'{len(to_unlink)} link(s) cannot be removed from collection: {col_name}')

        if len(to_remove) == 0:
            return

        force = kwargs.get('force') is True
        if not force:
            answer = prompt_utils.user_choice(
                f'Do you want to remove {len(to_remove)} link(s)?', 'yN', default='n')
            if answer.lower() not in ('y', 'yes'):
                return

        status = plugin.set_collection_links(unset_links=to_remove)
        if len(status.remove) > 0:
            for st in status.remove:
                if isinstance(st, admin_types.LinkStatus) and not st.success:
                    raise base.CommandError(f'Failed to unset collection link "{st.object_uid}": {st.message}')


class PedmCollectionListCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='list', description='List PEDM collections',
                                         parents=[base.report_output_parser])
        parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                            help='print verbose information')
        parser.add_argument('--type', dest='type', action='store', type=int,
                            help='collection type filter')
        parser.add_argument('--pattern', dest='pattern', action='store',
                            help='collection search pattern')

        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> Any:
        plugin = admin_plugin.get_pedm_plugin(context)

        table: List[List[Any]] = []
        row: List[Any]
        collection_type: Optional[int] = kwargs.get('type')
        verbose = kwargs.get('verbose') is True
        pattern = kwargs.get('pattern')

        if isinstance(collection_type, int):
            col_dict: Dict[str, List[admin_storage.PedmStorageCollectionLink]] = {}
            for col in plugin.collections.get_all_entities():
                if col.collection_type != collection_type:
                    continue
                col_dict[col.collection_uid] = list(plugin.storage.collection_links.get_links_for_subject(col.collection_uid))

            headers = ['collection_uid', 'value']
            if verbose:
                headers.extend(['link_info'])
            else:
                headers.extend(['link_count'])
            for (collection_uid, links) in col_dict.items():
                collection = plugin.collections.get_entity(collection_uid)
                if not collection:
                    continue
                cv = [f'{k}={v}' for k, v in collection.collection_data.items()]
                row = [collection_uid, cv]
                if verbose:
                    link_info = [f'{x.link_uid} ({pedm_shared.collection_link_type_to_name(x.link_type)})' for x in links]
                    row.append(link_info)
                else:
                    row.append(len(links))
                table.append(row)
        else:
            type_dict: Dict[int, List[admin_types.PedmCollection]] = {}
            for col in plugin.collections.get_all_entities():
                if col.collection_type not in type_dict:
                    type_dict[col.collection_type] = []
                type_dict[col.collection_type].append(col)

            headers = ['id', 'collection_type']
            if verbose:
                headers.extend(['collection_uid', 'value'])
            else:
                headers.extend(['value_count'])

            for (col_type, collections) in type_dict.items():
                col_type_name = pedm_shared.collection_type_to_name(col_type)
                if verbose:
                    for collection in collections:
                        cv = [f'{k}={v}' for k, v in collection.collection_data.items()]
                        table.append([col_type, col_type_name, collection.collection_uid, cv])
                else:
                    table.append([col_type, col_type_name, len(collections)])

        regex: Optional[Pattern[str]] = re.compile(fnmatch.translate(f'*{pattern}*')) if pattern else None
        if regex is not None:
            def any_match(row: Any) -> bool:
                if not row:
                    return False
                if not isinstance(row, list):
                    return False

                match = False
                for column in row:
                    column_values = []
                    if isinstance(column, list):
                        column_values.extend([x for x in column if isinstance(x, str)])
                    elif isinstance(column, str):
                        column_values.append(column)
                    match = any((True for x in column_values if regex.match(x)))
                    if match:
                        break
                return match

            table = [x for x in table if any_match(x)]

        table.sort(key=lambda x: x[0])
        fmt = kwargs.get('format')
        if fmt != 'json':
            headers = [report_utils.field_to_title(x) for x in headers]
        return report_utils.dump_report_data(table, headers, column_width=80, fmt=fmt, filename=kwargs.get('output'))


class PedmCollectionViewCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='view', description='Show PEDM collection details',
                                         parents=[base.report_output_parser])
        parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                            help='print verbose information')
        parser.add_argument('--link', dest='link', action='append', help='Show link details')
        parser.add_argument('collection',  nargs='+', help='Collection UID')

        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> Any:
        plugin = admin_plugin.get_pedm_plugin(context)

        collection_uid = kwargs.get('collection')
        if isinstance(collection_uid, str):
            collection_uid = [collection_uid]
        if not collection_uid:
            return

        collections: Dict[str, admin_types.PedmCollection] = {}

        for uid in collection_uid:
            coll = plugin.collections.get_entity(uid)
            if coll:
                collections[uid] = coll

        link_info: List[str] = []
        agent_link_data: Dict[Tuple[str, str], Dict[str, Any]] = {}
        link = kwargs.get('link')
        if isinstance(link, str):
            link = [link]
        if isinstance(link, list) and len(link) > 0:
            links: List[admin_types.CollectionLink] = []
            for c_uid in collection_uid:
                for l_uid in link:
                    cl = plugin.storage.collection_links.get_link(c_uid, l_uid)
                    if cl:
                        links.append(admin_types.CollectionLink(
                            collection_uid=c_uid, link_type=cl.link_type, link_uid=l_uid))
            if len(links) > 0:
                for cld in plugin.get_collection_links(links=links):
                    if not cld.link_data:
                        continue
                    collection_uid = cld.collection_link.collection_uid
                    link_uid = cld.collection_link.link_uid
                    try:
                        agent_data = json.loads(crypto.decrypt_aes_v2(cld.link_data, plugin.agent_key))
                        agent_link_data[(collection_uid, link_uid)] = agent_data
                    except:
                        pass
                link_info = list({x[1] for x in agent_link_data.keys()})

        verbose = kwargs.get('verbose') is True
        headers = ['collection_uid', 'collection_type', 'collection_value']
        if len(link_info) > 0:
            headers.extend((f'"{x}"' for x in link_info))
        else:
            if verbose:
                headers.append('link_uid')
            else:
                headers.append('link_count')
        table = []
        row: List[Any]
        for collection_uid, coll in collections.items():
            row = [collection_uid]
            if coll:
                collection_type = f'{pedm_shared.collection_type_to_name(coll.collection_type)} ({coll.collection_type})'
                row.append(collection_type)
                collection_value = [f'{k}={v}' for k, v in coll.collection_data.items()]
                row.append(collection_value)

                if len(link_info) > 0:
                    for link in link_info:
                        ld = agent_link_data[(coll.collection_uid, link)]
                        if ld:
                            row.append([f'{x[0]}={x[1]}' for x in ld.items()])
                        else:
                            row.append(None)
                else:
                    link_titles = list((f'{x.link_uid} ({pedm_shared.collection_link_type_to_name(x.link_type)})'
                                  for x in plugin.storage.collection_links.get_links_for_subject(collection_uid)))
                    if verbose:
                        row.append(link_titles)
                    else:
                        row.append(len(link_titles))
                table.append(row)

        table.sort(key=lambda x: x[0])
        fmt = kwargs.get('format')
        if fmt != 'json':
            headers = [report_utils.field_to_title(x) for x in headers]
        column_width = None if verbose else 50
        return report_utils.dump_report_data(table, headers, column_width=column_width, fmt=fmt, filename=kwargs.get('output'))


class PedmApprovalCommand(base.GroupCommandNew):
    def __init__(self):
        super().__init__('Manage PEDM approval requests and approvals')
        self.register_command_new(PedmApprovalListCommand(), 'list', 'l')
        self.register_command_new(PedmApprovalStatusCommand(), 'action', 'a')
        self.default_verb = 'list'


class PedmApprovalListCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='list', description='List PEDM approval requests',
                                         parents=[base.report_output_parser])
        parser.add_argument('--type', dest='type', action='store', choices=['approved', 'denied', 'pending'],
                            help='approval type filter')

        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> Any:
        plugin = admin_plugin.get_pedm_plugin(context)

        approval_type = kwargs.get('type')
        if isinstance(approval_type, str):
            approval_type = approval_type.lower()
        else:
            approval_type = None
        table: List[List[Any]] = []
        headers = ['approval_uid', 'approval_type', 'status', 'agent_uid', 'account_info', 'application_info', 'justification', 'expire_in', 'created']
        for approval in plugin.approvals.get_all_entities():
            approval_uid = approval.approval_uid
            a_status = plugin.storage.approval_status.get_entity(approval_uid)
            if a_status:
                if a_status.approval_status == NotificationCenter_pb2.NAS_APPROVED:
                    status = 'Approved'
                elif a_status.approval_status == NotificationCenter_pb2.NAS_DENIED:
                    status = 'Denied'
                elif a_status.approval_status == NotificationCenter_pb2.NAS_UNSPECIFIED:
                    status = 'Pending'
                else:
                    status = 'Unsupported'
            else:
                status = 'Pending'
            if approval_type and approval_type != status.lower():
                continue

            account_info = [y[:30] for y in (f'{k}={v}' for k, v in approval.account_info.items())]
            application_info = [y[:30] for y in (f'{k}={v}' for k, v in approval.application_info.items())]
            table.append([approval.approval_uid, pedm_shared.approval_type_to_name(approval.approval_type),
                          status, approval.agent_uid, account_info, application_info, approval.justification,
                          approval.expire_in, approval.created])

        table.sort(key=lambda x: x[8], reverse=True)
        fmt = kwargs.get('format')
        if fmt != 'json':
            headers = [report_utils.field_to_title(x) for x in headers]
        return report_utils.dump_report_data(table, headers, fmt=fmt, filename=kwargs.get('output'))


class PedmApprovalStatusCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='action', description='Modify PEDM approval requests')
        parser.add_argument('--approve', dest='approve', action='append',
                            help='Request UIDs for approval')
        parser.add_argument('--deny', dest='deny', action='append',
                            help='Request UIDs for denial')
        parser.add_argument('--remove', dest='remove', action='append',
                            help='Request UIDs for removal. UID, @approved, @denied, @pending')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> None:
        plugin = admin_plugin.get_pedm_plugin(context)

        logger = utils.get_logger()
        def verify_uid(uids: Any) -> Optional[List[bytes]]:
            if isinstance(uids, str):
                uids = [uids]
            if isinstance(uids, list):
                to_uid = []
                for uid in uids:
                    approve_uid = utils.base64_url_decode(uid)
                    if len(approve_uid) == 16:
                        to_uid.append(approve_uid)
                    else:
                        logger.warning(f'Invalid UID: {uid}')
                if len(to_uid) > 0:
                    return to_uid
            return None

        to_approve = verify_uid(kwargs.get('approve'))
        to_deny = verify_uid(kwargs.get('deny'))
        to_remove = kwargs.get('remove')
        if to_remove:
            if isinstance(to_remove, str):
                to_remove = [to_remove]
            to_remove_set: Set[bytes] = set()
            to_resolve = []
            for uid in to_remove:
                if uid == '@approved':
                    to_remove_set.update(
                        (utils.base64_url_decode(x.approval_uid) for x in plugin.storage.approval_status.get_all_entities() if x.approval_status == NotificationCenter_pb2.NAS_APPROVED))
                elif uid == '@denied':
                    to_remove_set.update(
                        (utils.base64_url_decode(x.approval_uid) for x in plugin.storage.approval_status.get_all_entities() if x.approval_status == NotificationCenter_pb2.NAS_DENIED))
                elif uid == '@pending':
                    to_remove_set.update(
                        (utils.base64_url_decode(x.approval_uid) for x in plugin.storage.approval_status.get_all_entities() if x.approval_status == NotificationCenter_pb2.NAS_UNSPECIFIED))
                else:
                    to_resolve.append(uid)
            if len(to_resolve) > 0:
                to_remove = verify_uid(to_resolve)
                if isinstance(to_remove, list):
                    to_remove_set.update(to_remove)
            to_remove = list(to_remove_set)

        status_rs = plugin.modify_approvals(to_approve=to_approve, to_deny=to_deny, to_remove=to_remove)
        if status_rs.add:
            for status in status_rs.add:
                if not status.success:
                    if isinstance(status, admin_types.EntityStatus):
                        logger.warning(f'Failed to approved "{status.entity_uid}": {status.message}')
        if status_rs.update:
            for status in status_rs.update:
                if not status.success:
                    if isinstance(status, admin_types.EntityStatus):
                        logger.warning(f'Failed to deny "{status.entity_uid}": {status.message}')
        if status_rs.remove:
            for status in status_rs.remove:
                if not status.success:
                    if isinstance(status, admin_types.EntityStatus):
                        logger.warning(f'Failed to remove "{status.entity_uid}": {status.message}')



from __future__ import annotations
import argparse
from ..discover import PAMGatewayActionDiscoverCommandBase, GatewayContext
from ...display import bcolors
from ... import vault
from ...discovery_common.user_service import UserService
from ...discovery_common.record_link import RecordLink
from ...discovery_common.constants import PAM_USER, PAM_MACHINE
from ...discovery_common.types import ServiceAcl
from ...keeper_dag.types import RefType, EdgeType
from ... import __version__
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ...vault import TypedRecord
    from ...params import KeeperParams


class PAMActionServiceAddCommand(PAMGatewayActionDiscoverCommandBase):
    parser = argparse.ArgumentParser(prog='pam-action-service-add')

    # The record to base everything on.
    parser.add_argument('--gateway', '-g', required=True, dest='gateway', action='store',
                        help='Gateway name or UID')

    parser.add_argument('--machine-uid', '-m', required=True, dest='machine_uid', action='store',
                        help='The UID of the Windows Machine record')
    parser.add_argument('--user-uid', '-u', required=True, dest='user_uid', action='store',
                        help='The UID of the User record')
    parser.add_argument('--type', '-t', required=True, choices=['service', 'task', 'iis'], dest='type',
                        action='store', help='Relationship to add [service, task, iis]')

    def get_parser(self):
        return PAMActionServiceAddCommand.parser

    def execute(self, params: KeeperParams, **kwargs):

        gateway = kwargs.get("gateway")
        machine_uid = kwargs.get("machine_uid")
        user_uid = kwargs.get("user_uid")
        rel_type = kwargs.get("type")

        print("")

        gateway_context = GatewayContext.from_gateway(params, gateway)
        if gateway_context is None:
            print(f"{bcolors.FAIL}Could not find the gateway configuration for {gateway}.")
            return

        if gateway_context is None:
            print(f"  {self._f('Cannot get gateway information. Gateway may not be up.')}")
            return

        user_service = UserService(record=gateway_context.configuration, params=params, fail_on_corrupt=False,
                                   agent=f"Cmdr/{__version__}")
        record_link = RecordLink(record=gateway_context.configuration, params=params, fail_on_corrupt=False,
                                 agent=f"Cmdr/{__version__}")

        ###############

        # Check to see if the record exists.
        machine_record = vault.KeeperRecord.load(params, machine_uid)  # type: Optional[TypedRecord]
        if machine_record is None:
            print(self._f("The machine record does not exists."))
            return

        # Make sure the record is a PAM Machine.
        if machine_record.record_type != PAM_MACHINE:
            print(self._f("The machine record is not a PAM Machine."))
            return

        # Make sure this machine is linked to the configuration record.
        machine_rl = record_link.get_record_link(machine_record.record_uid)
        if machine_rl is None:
            print(self._f("The machine record does not exists in the graph."))
            return

        # Edges from provider and machine might be wrong.
        # Should be a LINK edge, could be an ACL edge.
        if (machine_rl.get_edge(record_link.dag.get_root, edge_type=EdgeType.LINK) is None and
                machine_rl.get_edge(record_link.dag.get_root, edge_type=EdgeType.ACL) is None):
            print(self._f("The machine record does not belong to this gateway."))
            return

        ###############

        # Check to see if the record exists.
        user_record = vault.KeeperRecord.load(params, user_uid)  # type: Optional[TypedRecord]
        if user_record is None:
            print(self._f("The user record does not exists."))
            return

        # Make sure this user is a PAM User.
        if user_record.record_type != PAM_USER:
            print(self._f("The user record is not a PAM User."))
            return

        record_rotation = params.record_rotation_cache.get(user_record.record_uid)
        if record_rotation is not None:
            controller_uid = record_rotation.get("configuration_uid")
            if controller_uid is None or controller_uid != gateway_context.configuration_uid:
                print(self._f("The user record does not belong to this gateway. Cannot use this user."))
                return
        else:
            print(self._f("The user record does not have any rotation settings."))
            return

        ########

        # Make sure we are setting up a Windows machine.
        # Linux and Mac do not use passwords in services and cron jobs; no need to link.
        os_field = next((x for x in machine_record.fields if x.label == "operatingSystem"), None)
        if os_field is None:
            print(self._f("Cannot find the operating system field in this record."))
            return
        os_type = None
        if len(os_field.value) > 0:
            os_type = os_field.value[0]
        if os_type is None:
            print(self._f("The operating system field of the machine record is blank."))
            return
        if os_type != "windows":
            print(self._f("The operating system is not Windows. "
                          "PAM can only rotate the services and scheduled task password on Windows."))
            return

        # Get the machine service vertex.
        # If it doesn't exist, create one.
        machine_vertex = user_service.get_record_link(machine_record.record_uid)
        if machine_vertex is None:
            machine_vertex = user_service.dag.add_vertex(
                uid=machine_record.record_uid,
                name=machine_record.title,
                vertex_type=RefType.PAM_MACHINE)

        # Get the user service vertex.
        # If it doesn't exist, create one.
        user_vertex = user_service.get_record_link(user_record.record_uid)
        if user_vertex is None:
            user_vertex = user_service.dag.add_vertex(
                uid=user_record.record_uid,
                name=user_record.title,
                vertex_type=RefType.PAM_USER)

        # Get the existing service ACL and set the proper attribute.
        acl = user_service.get_acl(machine_vertex.uid, user_vertex.uid)
        if acl is None:
            acl = ServiceAcl()
        if rel_type == "service":
            acl.is_service = True
        elif rel_type == "task":
            acl.is_task = True
        else:
            acl.is_iis_pool = True

        # Make sure the machine has a LINK connection to the configuration.
        if not user_service.dag.get_root.has(machine_vertex):
            user_service.belongs_to(gateway_context.configuration_uid, machine_vertex.uid)

        # Add our new ACL edge between the machine and the yser.
        user_service.belongs_to(machine_vertex.uid, user_vertex.uid, acl=acl)

        user_service.save()

        if rel_type == "service":
            print(
                self._gr(
                    f"Success: Services running on this machine, using this user, will be updated and restarted after "
                    "password rotation."
                )
            )
        elif rel_type == "task":
            print(
                self._gr(
                    f"Success: Scheduled tasks running on this machine, using this user, will be updated after "
                    "password rotation."
                )
            )
        else:
            print(
                self._gr(
                    f"Success: IIS pools running on this machine, using this user, will be updated after "
                    "password rotation."
                )
            )



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper SDK
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from typing import Optional, List, Callable, Any

from .vault import TypedRecord, TypedField


class TypedRecordFacade:
    def __init__(self):   # type: () -> None
        self._record = None     # type: Optional[TypedRecord]

    def _get_record(self):
        return self._record

    def _set_record(self, record):
        if record is None or isinstance(record, TypedRecord):
            self._record = record
            self.load_typed_fields()
        else:
            raise ValueError('expected TypedRecord')

    def _get_title(self):
        if isinstance(self._record, TypedRecord):
            return self._record.title
        raise ValueError('typed record is not assigned')

    def _set_title(self, value):
        if isinstance(self._record, TypedRecord):
            self._record.title = value
        else:
            raise ValueError('typed record is not assigned')

    def _get_notes(self):
        if isinstance(self._record, TypedRecord):
            return self._record.notes
        raise ValueError('typed record is not assigned')

    def _set_notes(self, value):
        if isinstance(self._record, TypedRecord):
            self._record.notes = value
        else:
            raise ValueError('typed record is not assigned')

    record = property(fget=_get_record, fset=_set_record)
    title = property(fget=_get_title, fset=_set_title)
    notes = property(fget=_get_notes, fset=_set_notes)

    def load_typed_fields(self):
        pass


def string_list_getter(name):   # type: (str) -> Callable[[TypedRecordFacade], List[str]]
    def getter(obj):
        field = getattr(obj, name)
        if isinstance(field, TypedField):
            return field.value
    return getter


def string_getter(name):   # type: (str) -> Callable[[TypedRecordFacade], str]
    def getter(obj):
        field = getattr(obj, name)
        if isinstance(field, TypedField):
            return field.value[0] if len(field.value) > 0 else ''
    return getter


def string_setter(name):   # type: (str) -> Callable[[Any, str], None]
    def setter(obj, value):
        field = getattr(obj, name)
        if isinstance(field, TypedField):
            if value:
                if len(field.value) > 0:
                    field.value[0] = value
                else:
                    field.value.append(value)
            else:
                field.value.clear()
    return setter

def boolean_getter(name):   # type: (str) -> Callable[[TypedRecordFacade], bool]
    def getter(obj):
        field = getattr(obj, name)
        if isinstance(field, TypedField):
            value = field.value[0] if len(field.value) > 0 else None
            if value is None:
                return None
            elif isinstance(value, bool) is True:
                return value

            if str(value).lower() in ['true', 'yes', '1', 'on']:
                return True
            elif str(value).lower() in ['false', 'no', '0', 'off']:
                return False
            return None
    return getter

def boolean_setter(name):   # type: (str) -> Callable[[Any, str], None]
    def setter(obj, value):
        field = getattr(obj, name)
        if isinstance(field, TypedField):
            if value is not None:
                if isinstance(value, bool) is not True:
                    if str(value).lower() in ['true', 'yes', '1', 'on']:
                        value = True
                    elif str(value).lower() in ['false', 'no', '0', 'off']:
                        value = False
                if len(field.value) > 0:
                    field.value[0] = value
                else:
                    field.value.append(value)
            else:
                field.value.clear()
    return setter

def string_element_getter(name, element_name):   # type: (str, str) -> Callable[[Any], str]
    def getter(obj):
        field = getattr(obj, name)
        if isinstance(field, TypedField):
            if len(field.value) > 0 and isinstance(field.value[0], dict):
                return field.value[0].get(element_name) or ''
    return getter


def string_element_setter(name, element_name):   # type: (str, str) -> Callable[[Any, str], None]
    def setter(obj, value):
        field = getattr(obj, name)
        if isinstance(field, TypedField):
            if value:
                if len(field.value) == 0:
                    field.value.append({})
                if isinstance(field.value[0], dict):
                    field.value[0][element_name] = value
            else:
                if len(field.value) > 0 and isinstance(field.value[0], dict):
                    field.value[0][element_name] = value
    return setter


def list_element_getter(name, element_name):   # type: (str, str) -> Callable[[TypedRecordFacade], list]
    def getter(obj):
        field = getattr(obj, name)
        if isinstance(field, TypedField):
            if len(field.value) > 0 and isinstance(field.value[0], dict):
                if element_name not in field.value[0] or not isinstance(field.value[0][element_name], list):
                    field.value[0][element_name] = []
                return field.value[0][element_name]
    return getter


class FileRefRecordFacade(TypedRecordFacade):
    _file_ref_getter = string_list_getter('_file_ref')

    def __init__(self):
        super(FileRefRecordFacade, self).__init__()
        self._file_ref = None       # type: Optional[TypedField]

    def load_typed_fields(self):
        if self.record:
            self._file_ref = next((x for x in self.record.fields if x.type == 'fileRef'), None)
            if self._file_ref is None:
                self._file_ref = TypedField.new_field('fileRef', [])
                self.record.fields.append(self._file_ref)
        else:
            self._file_ref = None
        super(FileRefRecordFacade, self).load_typed_fields()

    @property
    def file_ref(self):
        return FileRefRecordFacade._file_ref_getter(self)


class LoginRecordFacade(FileRefRecordFacade):
    _login_getter = string_getter('_login')
    _login_setter = string_setter('_login')
    _password_getter = string_getter('_password')
    _password_setter = string_setter('_password')
    _url_getter = string_getter('_url')
    _url_setter = string_setter('_url')
    _oneTimeCode_getter = string_getter('_oneTimeCode')
    _oneTimeCode_setter = string_setter('_oneTimeCode')

    def __init__(self):
        super(LoginRecordFacade, self).__init__()
        self._login = None        # type: Optional[TypedField]
        self._password = None     # type: Optional[TypedField]
        self._url = None          # type: Optional[TypedField]
        self._oneTimeCode = None  # type: Optional[TypedField]

    @property
    def login(self):
        return LoginRecordFacade._login_getter(self)

    @login.setter
    def login(self, value):
        LoginRecordFacade._login_setter(self, value)

    @property
    def password(self):
        return LoginRecordFacade._password_getter(self)

    @password.setter
    def password(self, value):
        LoginRecordFacade._password_setter(self, value)

    @property
    def url(self):
        return LoginRecordFacade._url_getter(self)

    @url.setter
    def url(self, value):
        LoginRecordFacade._url_setter(self, value)

    @property
    def oneTimeCode(self):
        return LoginRecordFacade._oneTimeCode_getter(self)

    @oneTimeCode.setter
    def oneTimeCode(self, value):
        LoginRecordFacade._oneTimeCode_setter(self, value)

    def load_typed_fields(self):
        if self.record:
            self.record.type_name = 'login'
            self._login = next((x for x in self.record.fields if x.type == 'login'), None)
            if self._login is None:
                self._login = TypedField.new_field('login', '')
                self.record.fields.append(self._login)
            self._password = next((x for x in self.record.fields if x.type == 'password'), None)
            if self._password is None:
                self._password = TypedField.new_field('password', '')
                self.record.fields.append(self._password)
            self._url = next((x for x in self.record.fields if x.type == 'url'), None)
            if self._url is None:
                self._url = TypedField.new_field('url', '')
                self.record.fields.append(self._url)
            self._oneTimeCode = next((x for x in self.record.fields if x.type == 'oneTimeCode'), None)
            if self._oneTimeCode is None:
                self._oneTimeCode = TypedField.new_field('oneTimeCode', '')
                self.record.fields.append(self._oneTimeCode)
        else:
            self._login = None
            self._password = None
            self._url = None
            self._oneTimeCode = None

        super(LoginRecordFacade, self).load_typed_fields()



from data_vault import VaultEnvironment


class KeeperApiHelper:
    _expected_commands = []
    _vault_env = VaultEnvironment()

    @staticmethod
    def communicate_expect(actions):
        # type: (list) -> None
        KeeperApiHelper._expected_commands.clear()
        KeeperApiHelper._expected_commands.extend(actions)

    @staticmethod
    def is_expect_empty():
        # type: () -> bool
        return len(KeeperApiHelper._expected_commands) == 0

    @staticmethod
    def communicate_command(_, request):
        # type: (any, dict) -> dict
        rs = {
            'result': 'success',
            'result_code': '',
            'message': ''
        }
        action = KeeperApiHelper._expected_commands.pop(0)

        if callable(action):
            props = action(request)
            if type(props) == dict:
                rs.update(props)
            return rs

        if type(action) == str:
            if action == request['command']:
                return rs

        raise Exception()



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: commander@keepersecurity.com
#
# Example code to retrieve the password for a record
# stored in the vault.  This example also pulls configuration
# from config.json or writes the config file if it does not exist.
#
# Usage:
#    python update_password.py

import getpass
import os

from keepercommander import api, vault_extensions, generator
from keepercommander.__main__ import get_params_from_config
from keepercommander.commands.record_edit import RecordUpdateCommand

my_params = get_params_from_config(os.path.join(os.path.dirname(__file__), 'config.json'))

while not my_params.user:
    my_params.user = getpass.getpass(prompt='User(Email): ', stream=None)

api.login(my_params)
if not my_params.session_token:
    exit(1)
api.sync_down(my_params)

record_uid = ''

# Find any login record

# 1. direct API
record_uid = next((x.record_uid for x in vault_extensions.find_records(my_params, record_type='login')), None)

# 2. list command
# list_command = RecordListCommand()
# login_records = list_command.execute(my_params, format='json', type='login')
# record_list = json.loads(login_records)
# if isinstance(record_list, list) and len(record_list) > 0:
#     record_uid = record_list[0]['record_uid']

if record_uid:
    # 1. record-update command
    edit_command = RecordUpdateCommand()
    edit_command.execute(my_params, record=record_uid, fields=['password=$GEN'])

    # 2. direct API
    #    record = vault.KeeperRecord.load(my_params, record_uid)
    #    if isinstance(record, vault.TypedRecord):
    #        password = generator.generate(20)
    #        password_field = record.get_typed_field('password')
    #        if password_field:
    #            password_field.value = [password]
    #            record_management.update_record(my_params, record)
    api.sync_down(my_params)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from functools import wraps
from typing import Callable, Any
import logging
import sys, os, yaml
import re
from enum import Enum
from ... import utils

class LogLevel(Enum):
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    DEBUG = logging.DEBUG
    INFO = logging.INFO

class GlobalLogger:
    _instance = None
    _initialized = False

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(GlobalLogger, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        if not GlobalLogger._initialized:
            self._creat_logging_config()
            self._config = self._load_config()
            self._logger = logging.getLogger('keeper_service')
            self._setup_logger()
            GlobalLogger._initialized = True

    def _creat_logging_config(self):
        config_path = utils.get_default_path() / "logging_config.yaml"
    
        default_config = {
            "logging": {
                "enabled": True,
                "level": "INFO"  # Options: DEBUG, INFO, WARNING, ERROR
            }
        }
        if not os.path.exists(config_path):
        # Create the directory if it doesn't exist
            config_path.parent.mkdir(parents=True, exist_ok=True)
            # Write the default config
            with open(config_path, "w") as f:
                yaml.dump(default_config, f, sort_keys=False)
        return default_config["logging"]
    
    def _load_config(self):
        config_path = utils.get_default_path() / "logging_config.yaml";
        
        # config_path = os.getenv("LOGGING_CONFIG_PATH", "logging_config.yaml")
        if os.path.exists(config_path):
            with open(config_path, "r") as f:
                return yaml.safe_load(f).get("logging", {})
        return {"enabled": True, "level": "INFO"}
    
    def _setup_logger(self):
        if not self._config.get("enabled", True):
            # Disable logging if it's not enabled in the config
            logging.disable(logging.CRITICAL)
            return
        
        if not self._logger.handlers:
            handler = logging.StreamHandler(sys.stderr)
            formatter = logging.Formatter(
                '%(asctime)s [%(levelname)s] %(name)s - %(message)s',
                datefmt='%Y-%m-%d %H:%M:%S'
            )
            handler.setFormatter(formatter)
            self._logger.addHandler(handler)
            # Prevent log propagation to parent loggers to avoid duplicate entries
            self._logger.propagate = False
            # self._logger.setLevel(logging.INFO) # Change for debug
            log_level_str = self._config.get("level", "INFO").upper()
            log_level = getattr(logging, log_level_str, logging.INFO)
            self._logger.setLevel(log_level)

    def set_level(self, level: LogLevel):
        self._logger.setLevel(level.value)

    def info(self, message: str):
        self._logger.info(message)

    def debug(self, message: str):
        self._logger.debug(message)

    def warning(self, message: str):
        self._logger.warning(message)

    def error(self, message: str):
        self._logger.error(message)

    def exception(self, message: str):
        self._logger.exception(message)

def debug_decorator(fn: Callable) -> Callable:
    """Debug decorator - only active when logging level is DEBUG"""
    @wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if logger._logger.isEnabledFor(logging.DEBUG):
            args_repr = [repr(a) for a in args]
            kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
            signature = ", ".join(args_repr + kwargs_repr)
            logger.debug(f"Call: {fn.__name__}({signature})")
        
        value = fn(*args, **kwargs)
        
        if logger._logger.isEnabledFor(logging.INFO):
            logger.debug(f"Return: {fn.__name__} → {value!r}")
        
        return value
    return wrapper

def catch_all(fn: Callable) -> Callable:
    """Global exception handler"""
    @wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            logger.exception(f"Unhandled error in {fn.__name__}")
            raise
    return wrapper


def sanitize_debug_data(data: str) -> str:
    """Sanitize sensitive data from debug output."""
    if not data:
        return data
    
    sanitized = data
    
    # Sanitize common password patterns
    patterns = [
        (r'"password"\s*:\s*"[^"]*"', '"password": "***"'),
        (r'"login"\s*:\s*"[^"]*"', '"login": "***"'),  
        (r'"secret"\s*:\s*"[^"]*"', '"secret": "***"'),
        (r'"token"\s*:\s*"[^"]*"', '"token": "***"'),
        (r'"key"\s*:\s*"[^"]*"', '"key": "***"'),
        (r'password=[^\s]*', 'password=***'),
        (r'login=[^\s]*', 'login=***'),
        # Sanitize email addresses in logs to protect PII
        (r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b', '***@***.***'),
    ]
    
    for pattern, replacement in patterns:
        sanitized = re.sub(pattern, replacement, sanitized, flags=re.IGNORECASE)
    
    return sanitized


logger = GlobalLogger()


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com

import os

# Without mounted volumes, Lambda can only write to /tmp. 
# The following environment variables are needed to make sure the Python import cache 
# is written to /tmp and not the user folder.
os.environ['HOME'] = '/tmp'
os.environ['TMPDIR'] = '/tmp'
os.environ['TEMP'] = '/tmp'

from keepercommander import api
from keepercommander.__main__ import get_params_from_config

# By default, Keeper Commander will attempt to create a .keeper directory 
# in the user folder to store the JSON configuration.
# In this case we will create a .keeper directory in /tmp 
# to store the JSON configuration (using get_params_from_config()).
keeper_tmp = '/tmp/.keeper'
os.makedirs(keeper_tmp, exist_ok=True)

# ------------------------------------------------------
# Keeper initialization function
# ------------------------------------------------------
def get_params():
    # Change the default JSON configuration location to /tmp
    params = get_params_from_config(keeper_tmp + '/config.json') 

    # Set username and password for Keeper Commander login
    #params.config = {'sso_master_password': True} # Force Master-Password login for SSO users
    #params.server = os.environ.get('KEEPER_SERVER') # https://keepersecurity.com
    params.user = os.environ.get('KEEPER_USER')
    params.password = os.environ.get('KEEPER_PASSWORD')
    

    return params

# ------------------------------------------------------
# Keeper JSON report function
# ------------------------------------------------------
def get_keeper_report(params, kwargs):
    from keepercommander.commands.aram import AuditReportCommand
    from json import loads
    
    report_class = AuditReportCommand()
    report = report_class.execute(params, **kwargs)
    return loads(report)
    
# ------------------------------------------------------
# Keeper CLI function
# ------------------------------------------------------
def run_keeper_cli(params, command):
    from keepercommander import cli
    
    cli.do_command(params, command)
    # No return statement as this function runs the CLI command 
    # without returning anything in Python
    
# ------------------------------------------------------
# Lambda handler
# ------------------------------------------------------
def lambda_handler(event, context):
    # Initialize Keeper Commander params
    params = get_params()

    # Keeper login and sync
    api.login(params)
    api.sync_down(params)
    # Enterprise sync (for enterprise commands)
    api.query_enterprise(params)

    run_keeper_cli(
        params, 
        'device-approve -a'
    )
    
    run_keeper_cli(
        params, 
        'action-report --target locked --apply-action delete --dry-run'
    )

    return get_keeper_report(
        params,
        {
            'report_type':'raw', 
            'format':'json',
            'limit':100,
            'event_type':['login']
        }
    )



import logging
from typing import Dict, Any, List, Set, Callable

from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey

from .. import crypto
from ..proto import enterprise_pb2
import json

from .storage_types import StorageUser, StorageRecord, StorageTeam


class EnterpriseUser:
    def __init__(self):
        self.user_uid = 0
        self.full_name = ''
        self.email = ''
        self.status = not enterprise_pb2.OK
        self.job_title = ''
        self.node_id = 0
        self.records = set()
        self.active_records = set()
        self.trash_records = set()
        self.roles = []

    @staticmethod
    def load(entity, decrypt_fn):    # type: (StorageUser, Callable) -> EnterpriseUser
        user = EnterpriseUser()
        user.user_uid = entity.user_uid
        user.email = decrypt_fn(entity.email)
        user.full_name = decrypt_fn(entity.full_name)
        user.status = entity.status
        user.job_title = decrypt_fn(entity.job_title)
        user.node_id = entity.node_id
        return user


class Record:
    def __init__(self, data_source=None, ec_key=None):
        self.record_uid = ''
        self.record_uid_bytes = b''
        self.data = {}
        self.created = 0
        self.last_pw_change = 0
        self.last_modified = 0
        self.last_rotation = 0
        self.shared = False
        self.in_trash = False
        self.has_attachments = False
        self.user_permissions = dict()
        data_source and self.update_properties(data_source, ec_key)

    def update_properties(self, entity, ec_key):  # type: (StorageRecord, EllipticCurvePrivateKey) -> None
        def decrypt_data(encrypted, key):  # type: (bytes, EllipticCurvePrivateKey) -> Dict['str', Any]
            decrypted = {}
            try:
                data_json = crypto.decrypt_ec(encrypted, key) if encrypted else b'{}'
                decrypted = json.loads(data_json.decode())
            except:
                logging.debug('Cannot decrypt record \"%s\" info.', self.record_uid)
            return decrypted

        self.record_uid = entity.record_uid if not self.record_uid else self.record_uid
        self.record_uid_bytes = entity.record_uid_bytes if not self.record_uid_bytes else self.record_uid_bytes
        self.data = decrypt_data(entity.encrypted_data, ec_key) if not self.data else self.data
        self.shared = entity.shared
        self.in_trash = entity.in_trash
        self.has_attachments = entity.has_attachments


class UserRecord:
    def __init__(self, user_uid, record):   # type: (int, Record) -> None
        self.record = record
        self.user_uid = user_uid


class RecordPermissions:
    def __init__(self, record_uid, permissions):    # type: (str, int) -> None
        self.record_uid = record_uid
        self.permission_bits = permissions
        self.user_uid = -1

    @property
    def permissions(self):
        return RecordPermissions.to_permissions_str(self.permission_bits)

    @staticmethod
    def to_permissions_str(permission_bits):
        # type: (int) -> str
        permission_masks = {1: 'owner', 2: 'mask', 4: 'edit', 8: 'share', 16: 'share_admin'}
        permissions = [permission for mask, permission in permission_masks.items() if (permission_bits & mask)]
        if not permissions:
            permissions.append('read-only')
        return ','.join(permissions)


class Team:
    def __init__(self):
        self.team_uid = ''
        self.team_name = ''
        self.restrict_edit = True
        self.restrict_share = True
        self.users = []

    @staticmethod
    def load(entity):   # type: (StorageTeam) -> Team
        team = Team()
        team.team_uid = entity.team_uid
        team.team_name = entity.team_name
        team.restrict_edit = entity.restrict_edit
        team.restrict_share = entity.restrict_share
        return team


class SharedFolder:
    def __init__(self, folder_uid):
        self.folder_uid = folder_uid
        self.record_permissions = []    # type: List[RecordPermissions]
        self.users = set()              # type: Set[int]
        self.teams = set()              # type: Set[str]

    def update_record_permissions(self, permissions):  # type: (RecordPermissions) -> None
        if not any([p for p in self.record_permissions if p.record_uid == permissions.record_uid]):
            self.record_permissions.append(permissions)



from __future__ import annotations
from .edge import DAGEdge
from .types import EdgeType, RefType
from .crypto import generate_random_bytes, generate_uid_str, urlsafe_str_to_bytes
from .exceptions import DAGDeletionException, DAGIllegalEdgeException, DAGVertexException, DAGKeyException
from typing import Optional, Union, List, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from .dag import DAG
    Content = Union[str, bytes, dict]
    QueryValue = Union[list, dict, str, float, int, bool]
    import pydantic
    from pydantic import BaseModel


class DAGVertex:

    def __init__(self, dag: DAG, uid: Optional[str] = None, name: Optional[str] = None,
                 keychain: Optional[bytes] = None, vertex_type: RefType = RefType.GENERAL):

        self.dag = dag

        # If the UID is not set, generate a UID.
        if uid is None:
            uid = generate_uid_str()
        # Else verify that the UID is valid. The UID should be a 16-byte value that is web-safe base64 serialized.
        else:
            if len(uid) != 22:
                raise ValueError(f"The uid {uid} is not a 22 characters in length.")
            try:
                b = urlsafe_str_to_bytes(uid)
                if len(b) != 16:
                    raise ValueError("not 16 bytes")
            except Exception:
                raise ValueError("The uid does not appear to be web-safe base64 string contains a 16 bytes value.")

        # If the UID is the root UID, make sure the vertex type is not general.
        # The root vertex needs to be either PAM_NETWORK or PAM_USER, if not set to PAM_NETWORK.
        if uid == self.dag.uid and (vertex_type != RefType.PAM_NETWORK and vertex_type != RefType.PAM_USER):
            vertex_type = RefType.PAM_NETWORK
        self.vertex_type = vertex_type

        # If the name is not defined, use the UID. Name is not persistent in the DAG.
        # If you load the DAG, the web service will not return the name.
        if name is None:
            name = uid

        self._uid = uid
        self._name = name

        # The keychain is a list of keys that can be used.
        # The keychain may contain multiple keys, when loading the default graph (graph_id)
        # For normal editing, the keychain will contain only one key.
        self._keychain = []
        if keychain is not None:
            if isinstance(keychain, list) is False:
                keychain = [keychain]
            self._keychain += keychain

        # Is the keychain corrupt?
        self.corrupt = False

        # These are edges to which vertex own this vertex. This vertex belongs to. So this would
        self.edges: list[Optional[DAGEdge]] = []
        self.has_uid = []

        # Flag indicating that this vertex is active.
        # This means this vertex has an active edge connected to another vertex.
        self.active = True

        # By default, we will save this vertex; not skip_save.
        # If in the process building the graph, it is decided that a vertex should not be saved; this can be set to
        #  prevent the vertex from being saved.
        self._skip_save = False

    def __str__(self):
        ret = f"Vertex {self.uid}\n"
        ret += f"  python instance id: {id(self)}\n"
        ret += f"  name: {self.name}\n"
        ret += f"  keychain: {self.keychain}\n"
        ret += f"  active: {self.active}\n"
        ret += f"  edges:\n"
        for edge in self.edges:
            ret += f"    * type {self.dag.__class__.EDGE_LABEL.get(edge.edge_type)}"
            ret += f", connect to {edge.head_uid}"
            ret += f", path {edge.path}, "
            ret += f", active: {edge.active}"
            ret += f", modified: {edge.modified}"
            ret += f", content: {'yes' if edge.content is not None else 'no'}"
            ret += f", content type: {type(edge.content)}"
            ret += "\n"
        return ret

    def __repr__(self):
        return f"<DAGVertex {self.uid}>"

    def debug(self, msg: str, level: int = 0):
        self.dag.debug(msg, level=level)

    @property
    def name(self) -> str:
        """
        Get the name for vertex

        If the name is not defined, the UID will be returned.
        The name is not persistent.
        If loading a DAG, the name will not be set.

        :return:
        """
        if self._name is not None:
            return self._name
        return self._uid

    @property
    def key(self) -> Optional[Union[str, bytes]]:
        """
        Get a single key from the keychain.

        :return:
        """

        keychain = self.keychain
        if len(keychain) > 0:
            return self.keychain[0]

        return None

    @property
    def skip_save(self):
        return self._skip_save

    @skip_save.setter
    def skip_save(self, value):
        self._skip_save = value

        for vertex in self.has_vertices():
            vertex._skip_save = value

    def add_to_keychain(self, key: Union[str, bytes]):
        """
        Add a key to the keychain

        :param key: A decrypted key bytes or encrypted key str
        :return:
        """
        if key not in self._keychain:
            self._keychain.append(key)

    @property
    def keychain(self) -> Optional[List[Union[str, bytes]]]:
        """
        Get the keychain for the vertex.

        The key is stored on the edges, however, the key belongs to the vertex.
        KEY and ACL edges from this vertex will have the same encrypted key.
        It is simpler to store the key on the DAGVertex instance.

        The keychain in an array of keys.
        When using graph_id = 0, different graphs that have the same UID will
        have different keys.
        When decrypting DATA edges, each key in the keychain will be tried.

        If the keychain has not been set, check if any edges exist that require a key.
        If there are, then generate a random key.
        The load process will populate the key.
        If the vertex does not have a key in the keychain, it is because this is a newly
        added vertex.

        If there are no edges that require a key, then return None.
        """

        # If the vertex is root, then the keychain will be the key bytes.
        if self.dag.get_root == self:
            self._keychain = [self.dag.key]

        # If the keychain is empty, generate a key for a specific edge type.
        elif len(self._keychain) == 0:
            for e in self.edges:
                if e.edge_type in [EdgeType.KEY, EdgeType.DATA]:
                    self._keychain.append(generate_random_bytes(self.dag.__class__.UID_KEY_BYTES_SIZE))
                    break

        return self._keychain

    @keychain.setter
    def keychain(self, value: List[Union[str, bytes]]):
        """
        Set the key in the vertex.

        The save method will use this key for any KEY/ACL edges.
        A key of str type means it is encrypted.
        """
        self._keychain = value

    @property
    def has_decrypted_keys(self) -> Optional[bool]:
        """
        Does the vertex have a decrypted keys?

        If the vertex contains a KEY, ACL or DATA edge and if the key is bytes, then the key is decrypted.
        If it is a str type, then it is encrypted.
        """
        if len(self._keychain) > 0:
            for e in self.edges:
                if e.edge_type in [EdgeType.KEY, EdgeType.DATA]:
                    all_decrypted = True
                    for key in self._keychain:
                        if isinstance(key, bytes) is False:
                            all_decrypted = False
                            break
                    return all_decrypted
        return None

    @property
    def uid(self):
        """
        Get the vertex UID.

        Once set, don't allow it to be changed.
        """
        return self._uid

    def get_edge(self, vertex: DAGVertex, edge_type: EdgeType) -> DAGEdge:
        high_edge = None
        high_version = -1
        for edge in self.edges:
            # Get all the edge point at the same vertex.
            # Don't include DATA edges.
            if edge.head_uid == vertex.uid and edge.edge_type == edge_type:
                if edge.version > high_version:
                    high_version = edge.version
                    high_edge = edge
        return high_edge

    def get_highest_edge_version(self, head_uid: str) -> (int, Optional[DAGEdge]):
        """
        Find the highest DAGEdge version of all edge types.

        :param head_uid:
        :return:
        """

        high_edge = None
        high_version = -1
        for edge in self.edges:
            # Get all the edge point at the same vertex.
            # Don't include DATA edges.
            if edge.head_uid == head_uid:
                if edge.version > high_version:
                    high_edge = edge
                    high_version = edge.version
        return high_version, high_edge

    def edge_count(self, vertex: DAGVertex, edge_type: EdgeType) -> int:
        """
        Get the number of edges between two vertices.

        :param vertex:
        :param edge_type:
        :return:
        """
        count = 0
        for edge in self.edges:
            if edge.head_uid == vertex.uid and edge.edge_type == edge_type:
                count += 1
        return count

    def edge_by_type(self, vertex: DAGVertex, edge_type: EdgeType) -> List[DAGEdge]:
        edge_list = []
        for edge in self.edges:
            if edge.edge_type == edge_type and edge.head_uid == vertex.uid:
                edge_list.append(edge)
        return edge_list

    @property
    def has_data(self) -> bool:

        """
        Does this vertex contain a DATA edge?

        :return: True if vertex has a DATA edge.
        """

        for item in self.edges:
            if item.edge_type == EdgeType.DATA:
                return True
        return False

    def get_data(self, index: Optional[int] = None) -> Optional[DAGEdge]:
        """
        Get data edge

        If the index is None or 0, the latest data edge will be returned.
        A positive and negative, non-zero, index will return the same data.
        It will be the absolute value of the index from the latest data.
        This means the 1 or -1 will return the prior data.

        If there is no data, None is returned.

        :param index:
        :return:
        """

        data_list = self.edge_by_type(self, EdgeType.DATA)
        data_count = len(data_list)
        if data_count == 0:
            return None

        # If the index is None, get the latest.
        if index is None or index == 0:
            index = -1
        # Since -1 is the current, switch index to a negative number and subtract one more.
        # For example, 1 means prior, -1 would be the latest, so we need to subtract one to get -2.
        elif index > 0:
            index *= -1
            index -= 1
        # If already a negative index, just subtract one.
        else:
            index -= 1

        try:
            data = data_list[index]
        except IndexError:
            raise ValueError(f"The index is not valid. Currently there are {data_count} data edges")

        return data

    def add_data(self, content: Any, path: Optional[str] = None, modified: bool = True,
                 from_load: bool = False, needs_encryption: bool = True):

        """
        Add a DATA edge to the vertex.

        :param content: The content to store in the DATA edge.
        :param path: Simple string tag to identify the edge.
        :param modified: Does this modify the content?
                         By default, adding a DATA edge will flag that the edge has been modified.
                         If loading, modified will be set to False.
        :param from_load: This call is being performed the load() method.
                          Do not validate adding data.
        :param needs_encryption: Default is True.
                                 Does the content need to be encrypted?
        """

        self.debug(f"connect {self.uid} to DATA edge", level=1)

        # Are we trying to add DATA to a deleted vertex?

        if self.active is False:
            # If deleted, there will not be a KEY to decrypt the data.
            # Throw an exception if not from the loading method.
            if from_load is False:
                raise DAGDeletionException("This vertex is not active. Cannot add DATA edge.")
            # If from loading, do not add and do not throw an exception.
            return

        # Make sure the vertex belongs before auto saving. If it does not belong, it's just an orphan right now.
        # This only is checked if using this module is used to create the graph.
        if self.belongs_to_a_vertex is False and from_load is False:
            raise DAGVertexException(f"Before adding data, connect this vertex {self.uid} to another vertex.")

        # Make sure that we have a KEY.
        # Allow a DATA edge to be connected to the root vertex, which will not have a KEY edge.
        # Or if we are loading, allow out of sync edges.

        if needs_encryption is True:
            found_key_edge = self.dag.get_root == self or from_load is True
            if found_key_edge is False:
                for edge in self.edges:
                    if edge.edge_type == EdgeType.KEY:
                        found_key_edge = True
            if found_key_edge is False:
                raise DAGKeyException(f"Cannot add DATA edge without a KEY edge for vertex {self.uid}.")

        # Get the prior data, set the version and inactive the prior data.
        version = 0
        prior_data = self.get_data()
        if prior_data is not None:
            version = prior_data.version + 1
            prior_data.active = False

        # The tail UID is the UID of the vertex. Since data loops back to the vertex, the head UID is the same.
        self.edges.append(
            DAGEdge(
                vertex=self,
                edge_type=EdgeType.DATA,
                head_uid=self.uid,
                version=version,
                content=content,
                path=path,
                modified=modified,
                from_load=from_load,
                needs_encryption=needs_encryption
            )
        )

        # If using a history level, we want to remove edges if we exceed the history level.
        # The history level is per edge type.
        # It's FIFO, so we will remove the first edge type if we exceed the history level.
        if self.dag.history_level > 0:
            data_count = self.data_count()
            while data_count > self.dag.history_level:
                for index in range(0, len(self.edges) - 1):
                    if self.edges[index].edge_type == EdgeType.DATA:
                        del self.edges[index]
                        data_count -= 1
                        break

        self.dag.do_auto_save()

    def data_count(self):
        return self.edge_count(self, EdgeType.DATA)

    def data_delete(self):

        # Get the DATA edge.
        # It will be a reference to itself.
        data_edge = self.get_edge(self, EdgeType.DATA)
        if data_edge is None:
            self.debug("cannot delete the data, no data edge exists.")

        data_edge.active = False

        self.belongs_to(
            vertex=self,
            edge_type=EdgeType.DELETION
        )
        self.debug(f"deleted data edge for {self.uid}")

    @property
    def latest_data_version(self):
        version = -1
        for edge in self.edges:
            if edge.edge_type == EdgeType.DATA and edge.version > version:
                version = edge.version
        return version

    @property
    def content(self) -> Optional[Union[str, bytes]]:
        """
        Get the content of the active DATA edge.

        If the content is a str, then the content is encrypted.
        """
        data_edge = self.get_data()
        if data_edge is None:
            return None
        return data_edge.content

    @property
    def content_as_dict(self) -> Optional[dict]:
        """
        Get the content from the active DATA edge as a dictionary.
        :return: Content as a dictionary.
        """
        data_edge = self.get_data()
        if data_edge is None:
            return None
        return data_edge.content_as_dict

    @property
    def content_as_str(self) -> Optional[str]:
        """
        Get the content from the active DATA edge as a str.
        :return: Content as a str.
        """

        data_edge = self.get_data()
        if data_edge is None:
            return None
        return data_edge.content_as_str

    def content_as_object(self, meta_class: pydantic._internal._model_construction.ModelMetaclass) -> (
            Optional)[BaseModel]:
        """
        Get the content as a pydantic based object.

        :param meta_class: The class to return
        :return:
        """
        data_edge = self.get_data()
        if data_edge is None:
            return None

        return data_edge.content_as_object(meta_class)

    @property
    def has_key(self) -> bool:

        """
        Does this vertex contain any KEY or ACL edges?

        :return: True if vertex has a KEY or ACL edge.
        """

        for item in self.edges:
            if item.edge_type == EdgeType.KEY:
                return True
        return False

    def belongs_to(self, vertex: DAGVertex, edge_type: EdgeType, content: Optional[Any] = None,
                   path: Optional[str] = None, modified: bool = True, from_load: bool = False):

        """
        Connect a vertex to another vertex (as the owner).

        This will create an edge between this vertex and the passed in vertex.
        The passed in vertex will own this vertex.

        If the edge_type is a KEY or ACL, data will be treated as a key. If a DATA edge already exists, the
        edge_type will be changed to a KEY, if not a KEY or ACL edge_type.

        :param vertex: The vertex has this vertex.
        :param edge_type: The edge type that connects the two vertices.
        :param content: Data to store as the edges content.
        :param path: Text tag for the edge.
        :param modified: Does adding this edge modify the stored DAG?
        :param from_load: Is being connected from load() method?
        :return:
        """

        self.dag.debug(f"connect {self.uid} to {vertex.uid} with edge type {edge_type.value}", level=1)  # DAG_DEBUG_LEVEL env var

        if vertex is None:
            raise ValueError("Vertex is blank.")
        if self.uid == self.dag.uid and not (edge_type == EdgeType.DATA or edge_type == EdgeType.DELETION):
            if from_load is False:
                raise DAGIllegalEdgeException(f"Cannot create edge to self for edge type {edge_type}.")
            self.dag.logger.debug(f"vertex {self.uid} , the root vertex, "
                           f"attempted to create '{edge_type.value}' edge to self, skipping.")
            return

        # Cannot make an edge to the same vertex, unless the edge type is a DELETION.
        # Normally an edge to self is a DATA type, use add_data for that.
        # A DELETION edge to self is allowed.
        # Just means the DATA edge is being deleted.
        if self.uid == vertex.uid and not (edge_type == EdgeType.DATA or edge_type == EdgeType.DELETION):
            if from_load is False:
                raise DAGIllegalEdgeException(f"Cannot create edge to self for edge type {edge_type}.")
            self.dag.logger.debug(f"vertex {self.uid} attempted to make '{edge_type.value}' to self, skipping.")
            return

        # Figure out what version of the edge we are.

        version, _ = self.get_highest_edge_version(head_uid=vertex.uid)

        # If the new edge is not DELETION
        if edge_type != EdgeType.DELETION:

            # Find the current active edge for this edge type to make it inactive.
            current_edge_by_type = self.get_edge(vertex, edge_type)
            if current_edge_by_type is not None:
                current_edge_by_type.active = False

            # If we are adding a non-DELETION edge, it will inactivate the DELETION edge.
            highest_deletion_edge = self.get_edge(vertex, EdgeType.DELETION)
            if highest_deletion_edge is not None:
                highest_deletion_edge.active = False

        # Should we activate the vertex again?
        if self.active is False:

            # If the vertex is already inactive, and we are trying to delete, return.
            if edge_type == EdgeType.DELETION:
                return

            self.dag.logger.debug(f"vertex {self.uid} was inactive; reactivating vertex.")
            self.active = True

        # Create and append a new DAGEdge instance.
        # Disable the auto saving after the content is changed since the edge has not been appended yet.
        # Once the edge is created, disable blocking auto save for content changes.
        edge = DAGEdge(
            vertex=self,
            edge_type=edge_type,
            head_uid=vertex.uid,
            version=version + 1,
            block_content_auto_save=True,
            content=content,
            path=path,
            modified=modified,
            from_load=from_load
        )
        edge.block_content_auto_save = False

        self.edges.append(edge)
        if self.uid not in vertex.has_uid:
            vertex.has_uid.append(self.uid)

        self.dag.do_auto_save()

    def belongs_to_root(self, edge_type: EdgeType, path: Optional[str] = None):

        """
        Connect the vertex to the root vertex.

        :param edge_type: The type of edge to use for the connection.
        :param path: Short tag for this edge.
        :return:
        """

        self.debug(f"connect {self.uid} to root", level=1)

        if self.uid == self.dag.uid:
            raise DAGIllegalEdgeException("Cannot create edge to self.")

        if self.active is False:
            raise DAGDeletionException("This vertex is not active. Cannot connect to root.")

        # We are adding the root, we can enable auto save now.
        # We can get the correct stream id with an edge to the root vertex.
        self.belongs_to(self.dag.get_root, edge_type=edge_type, path=path)

        self.dag.allow_auto_save = True
        self.dag.do_auto_save()

    def has_vertices(self, edge_type: Optional[EdgeType] = None, allow_inactive: bool = False,
                     allow_self_ref: bool = False) -> List[DAGVertex]:

        """
        Get a list of vertices that belong to this vertex.
        :return: List of DAGVertex
        """

        vertices = []
        for uid in self.has_uid:

            # This will remove DATA and DATA that have changed to DELETION edges.
            # Prevent looping.
            if uid == self.uid and allow_self_ref is False:
                continue

            vertex = self.dag.get_vertex(uid)
            if edge_type is not None:
                edge = vertex.get_edge(self, edge_type=edge_type)
                if edge is not None:
                    vertices.append(vertex)

            # If no edge type was specified, do not return DATA and DELETION.
            # Also do not include vertices that are inactive by default.
            elif edge_type != EdgeType.DATA and edge_type != EdgeType.DELETION:
                if vertex.active is True or allow_inactive is True:
                    vertices.append(vertex)

        return vertices

    def has(self, vertex: DAGVertex, edge_type: Optional[EdgeType] = None) -> bool:

        """
        Does this vertex have the passed in vertex?

        :return: True if request vertex belongs to this vertex.
                 False if it does not.
        """

        vertices = self.has_vertices(edge_type=edge_type)
        return vertex in vertices

    def belongs_to_vertices(self) -> List[DAGVertex]:
        """
        Get a list of vertices that this vertex belongs to
        :return:
        """

        vertices = []
        for edge in self.edges:
            # If the edge is not a DATA or DELETION type, and the edge is the highest version/active
            if edge.edge_type != EdgeType.DATA and edge.edge_type != EdgeType.DELETION and edge.active is True:

                # The head will point at the remote vertex.
                # If it is active, and not already in the list, add it to the list of vertices this vertex belongs to.
                vertex = self.dag.get_vertex(edge.head_uid)
                if vertex.active is True and vertex not in vertices:
                    vertices.append(vertex)
        return vertices

    @property
    def belongs_to_a_vertex(self) -> bool:
        """
        Does this vertex belong to another vertex?
        :return:
        """

        # If this is the root vertex, return True.
        # Where this is being called should handle operations involving the root vertex.
        if self.dag.get_root == self:
            return True

        return len(self.belongs_to_vertices()) > 0

    def disconnect_from(self, vertex: DAGVertex, path: Optional[str] = None):

        """
        Disconnect this vertex from another vertex.

        This will add a DELETION edge between two vertices.
        If the vertex no longer belongs to another vertex, the vertex will be deleted.

        :param vertex: The vertex this vertex belongs to
        :param path: an Optional path for the DELETION edge.
        :return:
        """

        if vertex is None:
            raise ValueError("Vertex is blank.")

        # Flag all the edges as inactive.
        for edge in self.edges:
            if edge.head_uid == vertex.uid and edge.edge_type:
                edge.active = False

        # Add the DELETION edge
        self.belongs_to(
            vertex=vertex,
            edge_type=EdgeType.DELETION,
            path=path
        )

        # If all the KEY edges are inactive now, the DATA edge needs to be made inactive.
        # There is no longer a KEY edge to decrypt the DATA.
        has_active_key_edge = False
        for edge in self.edges:
            if edge.edge_type == EdgeType.KEY and edge.active is True:
                has_active_key_edge = True
                break
        if has_active_key_edge is False:
            for edge in self.edges:
                if edge.edge_type == EdgeType.DATA:
                    edge.active = False

        if self.belongs_to_a_vertex is False:
            self.debug(f"vertex {self.uid} is now not active", level=1)
            self.active = False

    def delete(self, ignore_vertex: Optional[DAGVertex] = None):

        """
        Delete a vertex

        Deleting a vertex will inactivate the vertex.
        It will also inactivate any vertices, and their edges, that belong to the vertex.
        It will not inactivate a vertex that belongs to multiple vertices.
        :return:
        """

        def _delete(vertex, prior_vertex):

            # Do not delete the root vertex
            if vertex.uid == self.dag.uid:
                self.debug(f"  * vertex is root, cannot delete root", level=2)
                return

            self.debug(f"> checking vertex {vertex.uid}")

            # Should we ignore a vertex?
            # If deleting an edge, we want to ignore the vertex that owns the edge.
            # This prevents circular calls.
            if ignore_vertex is not None and vertex.uid == ignore_vertex.uid:
                return

            # Get a list of vertices that belong to this vertex (v)
            has_v = vertex.has_vertices()

            if len(has_v) > 0:
                self.debug(f"  * vertex has {len(has_v)} vertices that belong to it.", level=2)
                for v in has_v:
                    self.debug(f"    checking {v.uid}")
                    _delete(v, vertex)
            else:
                self.debug(f"  * vertex {vertex.uid} has NO vertices.", level=2)

            for e in list(vertex.edges):
                if e.edge_type != EdgeType.DATA and (prior_vertex is None or e.head_uid == prior_vertex.uid):
                    e.delete()
            if vertex.belongs_to_a_vertex is False:
                self.debug(f"  * inactive vertex {vertex.uid}")
                vertex.active = False

        self.debug(f"DELETING vertex {self.uid}", level=3)

        # Perform the DELETION edges save in one batch.
        # Get the current allowed auto save state, and disable auto save.
        current_allow_auto_save = self.dag.allow_auto_save
        self.dag.allow_auto_save = False

        _delete(self, None)

        # Restore the allow auto save and trigger auto save()
        self.dag.allow_auto_save = current_allow_auto_save
        self.dag.do_auto_save()

    def walk_down_path(self, path: Union[str, List[str]]) -> Optional[DAGVertex]:

        """
        Walk the vertices using the path and return the vertex starting at this vertex.

        :param path: An array of path string, or string where the path is joined with a "/" (i.e., think URL)
        :return: DAGVertex is the path completes, None is failure.
        """

        self.debug(f"walking path in vertex {self.uid}", level=2)

        # If the path is str, break it into an array. Get rid of leading /
        if isinstance(path, str) is True:
            self.debug("path is str, break into array", level=2)
            if path.startswith("/") is True:
                path = path[1:]
            path = path.split("/")

        # Unshift the path

        current_path = path[0]
        path = path[1:]
        self.debug(f"current path: {current_path}", level=2)
        self.debug(f"path left: {path}", level=2)

        # Check the DATA edges.
        # If a DATA edge has the current path, return this vertex.
        for edge in self.edges:
            if edge.edge_type != EdgeType.DATA:
                continue
            if edge.path == current_path:
                return self

        # Check the vertices that belong to this vertex for edges going to this vertex and the path matches.
        for vertex in self.has_vertices():
            self.debug(f"vertex {self.uid} has {vertex.uid}", level=2)
            for edge in vertex.edges:
                # If the edge matches the current path, the head of the edge is this vertex, a route exists.
                if edge.path == current_path and edge.head_uid == self.uid:
                    # If there is no path left, this is our vertex
                    if len(path) == 0:
                        return vertex
                    # If there is still more path, call vertex to walk more of the path.
                    else:
                        return vertex.walk_down_path(path)
        return None

    def get_paths(self) -> List[str]:
        """
        Get paths from this vertex to vertex owned by this vertex.
        :return: List of string paths
        """

        paths = []
        for vertex in self.has_vertices():
            for edge in vertex.edges:
                if edge.path is None or edge.path == "":
                    continue
                paths.append(edge.path)

        return paths



#!/usr/bin/env python3

"""Tests for keeper interactive shell autocompletion."""

import pytest

# This crufty import makes keepercommander.autocomplete import successfully.
import keepercommander.cli

import keepercommander.autocomplete as autocomplete

# And this silences the IDE/Editor warning from importing something we don't intend to use.
_ = keepercommander.cli


def check_roundtrip_w_in_double_quotes(string):
    """Test string for escape -> unescape, both with double quotes."""
    escaped = autocomplete.escape_string(True, string)
    unescaped = autocomplete.unescape_string(True, escaped)
    assert string == unescaped


def check_roundtrip_wo_in_double_quotes(string):
    """Test string for escape -> unescape, both without double quotes."""
    escaped = autocomplete.escape_string(False, string)
    unescaped = autocomplete.unescape_string(False, escaped)
    assert string == unescaped


strings_to_test = (
    '\\"k',
    'a//',
    'c ',
    'c\\ ',
    r"e\'f",
    r'g\"',
    'i\\',
    'k\\"',
    r'r\s',
    r'"\"k',
    '"a//',
    '"c ',
    '"c\\ ',
    '"e\'f',
    r'"g\"',
    r'"i\\',
    r'"k\"',
    r'"r\\',
    r"'",
    r"''",
    r"'a'",
)


@pytest.mark.parametrize('string_to_test', strings_to_test)
def test_w_double_quotes(string_to_test):
    r"""Test if a string round trips with in_double_quotes."""
    check_roundtrip_w_in_double_quotes(string_to_test)


@pytest.mark.parametrize('string_to_test', strings_to_test)
def test_wo_double_quotes(string_to_test):
    r"""Test if a string round trips without in_double_quotes."""
    check_roundtrip_wo_in_double_quotes(string_to_test)



# -*- coding: utf-8 -*-
#  _  __  
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|            
#
# Keeper Commander 
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#
import logging
import oracledb

"""Commander Plugin for Oracle Database Server
   Dependencies: 
       pip3 install oracledb
"""


ORACLE_MAX_PASSWORD_LENGTH = 30


class Rotator:
    def __init__(self, login, password, host='localhost', port=None, db='', dsn=None, **kwargs):
        self.host = host
        self.port = port
        self.login = login
        self.password = password
        self.db = db
        self.dsn = dsn

    def rotate_start_msg(self):
        """Display msg before starting rotation"""
        port_msg = f' and on port "{self.port}"' if self.port else ''
        if self.dsn:
            host_msg = f'on dsn "{self.dsn}"'
            db_msg = ''
        else:
            host_msg = f'on host "{self.host}"'
            db_msg = f' to connect to db "{self.db}"' if self.db else ''
        logging.info(
            f'Rotating with Oracle plugin {host_msg}{port_msg} using login "{self.login}"{db_msg}...'
        )

    @staticmethod
    def adjust(new_password):
        # Oracle password has a maximum length
        return new_password[:ORACLE_MAX_PASSWORD_LENGTH]

    def revert(self, record, new_password):
        """Revert rotation of an Oracle database password"""
        self.rotate(record, new_password, revert=True)

    def rotate(self, record, new_password, revert=False):
        """Rotate an Oracle database password"""
        if revert:
            old_password = new_password
            new_password = self.password
        else:
            old_password = self.password

        user = self.login
        dsn = self.dsn if self.dsn else f'{self.host}/{self.db}'
        kwargs = {'user': user, 'password': old_password, 'dsn': dsn}
        if self.port:
            kwargs['port'] = self.port

        connection = ''
        result = False
        try:
            connection = oracledb.connect(**kwargs)
            with connection.cursor() as cursor:
                logging.debug(f'Connected to {dsn}')
                sql = f'ALTER USER {user} IDENTIFIED BY "{new_password}" ACCOUNT UNLOCK'
                cursor.execute(sql)
            result = True
        except Exception as e:
            logging.error(f'Error during connection to Oracle server: {e}')
        finally:
            if connection:
                connection.close()
        return result
